"""Testing utilities for building expected queries for reviews.

Version Added:
    5.0.7
"""

from __future__ import annotations

from typing import Dict, Optional, Sequence, Set, TYPE_CHECKING, Union

from django.contrib.auth.models import AnonymousUser, User
from django.db.models import Q

from reviewboard.reviews.models import Review
from reviewboard.reviews.testing.queries.review_groups import \
    get_review_groups_accessible_ids_equeries
from reviewboard.scmtools.testing.queries import \
    get_repositories_accessible_ids_equeries
from reviewboard.site.models import LocalSite

if TYPE_CHECKING:
    from django_assert_queries.query_comparator import ExpectedQueries
    from typing_extensions import NotRequired, TypedDict, Unpack

    from reviewboard.site.models import AnyOrAllLocalSites
    from reviewboard.testing.queries.base import ExpectedQResult

    class _AccessibleKwargs(TypedDict):
        user: Union[AnonymousUser, User]
        local_site: NotRequired[AnyOrAllLocalSites]
        has_local_sites_in_db: NotRequired[bool]
        filter_private: NotRequired[bool]
        status: NotRequired[Optional[str]]
        public: NotRequired[Optional[bool]]
        base_reply_to: NotRequired[Optional[Review]]
        extra_query: NotRequired[Q]
        accessible_repository_ids: NotRequired[Sequence[int]]
        accessible_review_group_ids: NotRequired[Sequence[int]]
        has_view_invite_only_groups_perm: NotRequired[bool]
        needs_local_site_profile_query: NotRequired[bool]


##########################
# Q-expression utilities #
##########################

def get_reviews_accessible_q(
    *,
    user: Union[AnonymousUser, User],
    local_site: AnyOrAllLocalSites = None,
    has_local_sites_in_db: bool = False,
    filter_private: bool = False,
    status: Optional[str] = None,
    public: Optional[bool] = None,
    base_reply_to: Optional[Review] = None,
    extra_query: Q = Q(),
    accessible_repository_ids: Sequence[int] = [],
    accessible_review_group_ids: Sequence[int] = [],
    has_view_invite_only_groups_perm: bool = False,
    needs_local_site_profile_query: bool = False,
) -> ExpectedQResult:
    """Return a Q expression for accessible review queries.

    Version Added:
        5.0.7

    Args:
        user (django.contrib.auth.models.AnonymousUser or
              django.contrib.auth.models.User):
            The user that's checked for access.

        local_site (reviewboard.site.models.LocalSite, optional):
            The optional Local Site value used in the queries.

        has_local_sites_in_db (bool, optional):
            Whether the query expects Local Sites to be in the database.

        filter_private (bool, optional):
            Whether review requests would be filtered using ACL checks.

        status (str, optional):
            The optional status that would be used in the query.

        public (bool, optional):
            The optional public state to match in queries.

        base_reply_to (reviewboard.reviews.models.review.Review, optional):
            Any review that this results would be in reply to.

        extra_query (django.db.models.Q, optional):
            An extra query filter that would be included.

        accessible_repository_ids (list of int, optional):
            A list of accessible repository IDs that would be expected in the
            query.

        accessible_review_group_ids (list of int, optional):
            A list of accessible review group IDs that would be expected in the
            query.

        needs_local_site_profile_query (bool, optional):
            Whether the query should fetch a
            :py:class:`~reviewboard.accounts.models.LocalSiteProfile`.

            Set this to ``False`` if this should be cached at this point.

    Returns:
        dict:
        The expected Q results.
    """
    include_local_site_q = (has_local_sites_in_db and
                            local_site is not LocalSite.ALL)

    prep_equeries: ExpectedQueries = []

    is_authenticated: bool
    is_superuser: bool

    if user is None or isinstance(user, AnonymousUser):
        is_authenticated = False
        is_superuser = False
    else:
        is_authenticated = True
        is_superuser = user.is_superuser

        if filter_private and not is_superuser:
            prep_equeries += get_repositories_accessible_ids_equeries(
                user=user,
                local_site=local_site,
                visible_only=False)
            prep_equeries += get_review_groups_accessible_ids_equeries(
                user=user,
                local_site=local_site,
                visible_only=False,
                needs_local_site_profile_query=needs_local_site_profile_query)

    # This is intended to be verbose, to ensure we're matching exactly the
    # queries we expect. We want to minimize building of queries.
    #
    # That said, to keep this maintainable, query is built in 2 stages:
    #
    # 1. Initial caller-requested access control/visibility queries
    # 2. Access control filter queries
    init_q: Q
    access_q: Q
    tables: Set[str] = {'reviews_review'}
    join_types: Dict[str, str] = {}

    # Stage 1: Initial filtering Q-expression.
    if status:
        tables.add('reviews_reviewrequest')
        join_types['reviews_reviewrequest'] = 'INNER JOIN'

        if include_local_site_q:
            init_q = (Q(base_reply_to=base_reply_to) &
                      Q(review_request__status=status) &
                      Q(review_request__local_site=local_site) &
                      extra_query)
        else:
            init_q = (Q(base_reply_to=base_reply_to) &
                      Q(review_request__status=status) &
                      extra_query)
    else:
        if include_local_site_q:
            tables.add('reviews_reviewrequest')
            join_types['reviews_reviewrequest'] = 'INNER JOIN'

            init_q = (Q(base_reply_to=base_reply_to) &
                      Q(review_request__local_site=local_site) &
                      extra_query)
        else:
            init_q = (Q(base_reply_to=base_reply_to) &
                      extra_query)

    # Stage 2: Access control Q-expression.
    if filter_private:
        if is_superuser:
            if public is None:
                access_q = Q()
            else:
                access_q = Q(public=public)
        elif is_authenticated:
            if public is None:
                tables.update({
                    'reviews_reviewrequest',
                    'reviews_reviewrequest_target_groups',
                    'reviews_reviewrequest_target_people',
                })
                join_types.update({
                    'reviews_reviewrequest': 'INNER JOIN',
                    'reviews_reviewrequest_target_groups': 'LEFT OUTER JOIN',
                    'reviews_reviewrequest_target_people': 'LEFT OUTER JOIN',
                })

                access_q = (
                    Q(user=user) |
                    (Q(public=True) &
                     (Q(review_request__repository=None) |
                      Q(review_request__repository__in=(
                          accessible_repository_ids))) &
                     (Q(review_request__target_people=user) |
                      Q(review_request__target_groups=None) |
                      Q(review_request__target_groups__in=(
                          accessible_review_group_ids))))
                )
            elif public is True:
                tables.update({
                    'reviews_reviewrequest',
                    'reviews_reviewrequest_target_groups',
                    'reviews_reviewrequest_target_people',
                })
                join_types.update({
                    'reviews_reviewrequest': 'INNER JOIN',
                    'reviews_reviewrequest_target_groups': 'LEFT OUTER JOIN',
                    'reviews_reviewrequest_target_people': 'LEFT OUTER JOIN',
                })

                access_q = (
                    Q(public=True) &
                    (Q(review_request__repository=None) |
                     Q(review_request__repository__in=(
                         accessible_repository_ids))) &
                    (Q(review_request__target_people=user) |
                     Q(review_request__target_groups=None) |
                     Q(review_request__target_groups__in=(
                         accessible_review_group_ids)))
                )
            elif public is False:
                access_q = (Q(public=False) &
                            Q(user=user))
        else:
            assert user.is_anonymous
            assert public is not False, (
                "get_review_accessible_q can't be passed with public=False, "
                "filter_private, and anonymous users. ReviewManager._query "
                "supports this, but it cannot be expressed as a Q(). Make "
                "sure to check this condition manually in tests instead."
            )

            tables.update({
                'reviews_group',
                'reviews_reviewrequest',
                'reviews_reviewrequest_target_groups',
                'scmtools_repository',
            })
            join_types.update({
                'reviews_group': 'LEFT OUTER JOIN',
                'reviews_reviewrequest': 'INNER JOIN',
                'reviews_reviewrequest_target_groups': 'LEFT OUTER JOIN',
                'scmtools_repository': 'LEFT OUTER JOIN',
            })

            access_q = (
                (Q(review_request__repository=None) |
                 Q(review_request__repository__public=True)) &
                (Q(review_request__target_groups=None) |
                 Q(review_request__target_groups__invite_only=False)) &
                Q(public=True)
            )
    else:
        if public is None:
            access_q = Q()
        else:
            access_q = Q(public=public)

    return {
        'join_types': join_types,
        'prep_equeries': prep_equeries,
        'q': init_q & access_q,
        'tables': tables,
    }


def get_reviews_from_user_q(
    *,
    from_user: Union[str, User],
    **kwargs: Unpack[_AccessibleKwargs],
) -> ExpectedQResult:
    """Return a Q expression for accessible reviews from a user.

    This corresponds to a call from
    :py:meth:`ReviewRequest.objects.from_user()
    <reviewboard.reviews.managers.ReviewRequestManager.from_user>`.

    Version Added:
        5.0.7

    Args:
        from_user (str or django.contrib.auth.models.User):
            The username or user that the reviews will be from.

        **kwargs (dict, optional):
            Keyword arguments to pass to
            :py:func:`get_reviews_accessible_q`.

    Returns:
        list of dict:
        The list of expected queries.
    """
    extra_tables: Set[str] = set()
    extra_join_types: Dict[str, str] = {}

    if isinstance(from_user, User):
        extra_query = Q(user=from_user)
    else:
        assert isinstance(from_user, str)

        extra_query = Q(user__username=from_user)
        extra_tables.add('auth_user')
        extra_join_types['auth_user'] = 'INNER JOIN'

    kwargs['extra_query'] = extra_query

    q_result = get_reviews_accessible_q(**kwargs)
    q_result.setdefault('join_types', {}).update(extra_join_types)
    q_result['tables'].update(extra_tables)

    return q_result


############################
# Expected Query utilities #
############################

def get_reviews_accessible_equeries(
    *,
    distinct: bool = False,
    **kwargs: Unpack[_AccessibleKwargs],
) -> ExpectedQueries:
    """Return queries for accessible reviews.

    This corresponds to a call from
    :py:meth:`ReviewRequest.objects.accessible()
    <reviewboard.reviews.managers.ReviewRequestManager.accessible>`.

    Version Added:
        5.0.7

    Args:
        distinct (bool, optional):
            Whether this is a distinct query.

        **kwargs (dict, optional):
            Keyword arguments to pass to
            :py:func:`get_reviews_accessible_q`.

    Returns:
        list of dict:
        The list of expected queries.
    """
    kwargs.setdefault('filter_private', True)

    q_result = get_reviews_accessible_q(**kwargs)
    q_tables = q_result['tables']

    return q_result.get('prep_equeries', []) + [
        {
            '__note__': 'Fetch accessible reviews',
            'distinct': distinct,
            'join_types': q_result.get('join_types', {}),
            'num_joins': len(q_tables) - 1,
            'model': Review,
            'tables': q_tables,
            'where': q_result['q'],
        },
    ]


def get_reviews_from_user_equeries(
    *,
    from_user: Union[str, User],
    **kwargs: Unpack[_AccessibleKwargs],
) -> ExpectedQueries:
    """Return queries for accessible reviews from a user.

    This corresponds to a call from
    :py:meth:`ReviewRequest.objects.from_user()
    <reviewboard.reviews.managers.ReviewRequestManager.from_user>`.

    Version Added:
        5.0.7

    Args:
        from_user (str or django.contrib.auth.models.User):
            The username or user that the reviews will be from.

        **kwargs (dict, optional):
            Keyword arguments to pass to
            :py:func:`get_reviews_from_user_q`.

    Returns:
        list of dict:
        The list of expected queries.
    """
    q_result = get_reviews_from_user_q(from_user=from_user, **kwargs)
    q_tables = q_result['tables']

    if isinstance(from_user, User):
        username = from_user.username
    else:
        assert isinstance(from_user, str)

        username = from_user

    return q_result.get('prep_equeries', []) + [
        {
            '__note__': f'Fetch accessible reviews from user "{username}"',
            'distinct': False,
            'join_types': q_result.get('join_types', {}),
            'num_joins': len(q_tables) - 1,
            'model': Review,
            'tables': q_tables,
            'where': q_result['q'],
        },
    ]



"""Error definitions for SSO.

Version Added:
    5.0
"""


class InvalidUsernameError(ValueError):
    """Error for when a username is invalid.

    Version Added:
        5.0
    """


class BadSSOResponseError(ValueError):
    """Error for when we get a bad response from the SSO provider.

    Version Added:
        5.0
    """



"""The diff opcode generator."""

from __future__ import annotations

import logging
import os
import re
from typing import Optional, TYPE_CHECKING

from reviewboard.deprecation import RemovedInReviewBoard10_0Warning
from reviewboard.diffviewer.processors import (filter_interdiff_opcodes,
                                               post_process_filtered_equals)
from reviewboard.diffviewer.settings import DiffSettings

if TYPE_CHECKING:
    from collections.abc import Sequence

    from django.http import HttpRequest

    from reviewboard.diffviewer.differ import Differ


logger = logging.getLogger(__name__)


class MoveRange(object):
    """Stores information on a move range.

    This will store the start and end of the range, and all groups that
    are a part of it.
    """
    def __init__(self, start, end, groups=[]):
        self.start = start
        self.end = end
        self.groups = groups

    @property
    def last_group(self):
        return self.groups[-1]

    def add_group(self, group, group_index):
        if self.groups[-1] != group:
            self.groups.append((group, group_index))

    def __repr__(self):
        return '<MoveRange(%d, %d, %r)>' % (self.start, self.end, self.groups)


class DiffOpcodeGenerator:
    """The diff opcode generator."""

    ALPHANUM_RE = re.compile(r'\w')
    WHITESPACE_RE = re.compile(r'\s')

    MOVE_PREFERRED_MIN_LINES = 2
    MOVE_MIN_LINE_LENGTH = 20

    #: The default width for a tabstop.
    TAB_SIZE = DiffSettings.DEFAULT_TAB_SIZE

    ######################
    # Instance variables #
    ######################

    #: The raw contents for the diff.
    diff: Optional[bytes]

    #: The differ being used to generate the diff.
    differ: Differ

    #: The diff settings object.
    #:
    #: Version Added:
    #:     7.0.4
    diff_settings: DiffSettings

    #: The normalized lines of the original content of the filediff.
    #:
    #: This is used for interdiff filtering.
    #:
    #: Version Added:
    #:     8.0
    filediff_orig_lines: Sequence[str] | None

    #: The raw contents of the interdiff range diff.
    interdiff: Optional[bytes]

    #: The normalized lines of the original content of the interfilediff.
    #:
    #: This is used for interdiff filtering.
    #:
    #: Version Added:
    #:     8.0
    interfilediff_orig_lines: Sequence[str] | None

    #: The HTTP request from the client.
    request: Optional[HttpRequest]

    #: Whether to perform interdiff filtering.
    #:
    #: Version Added:
    #:     8.0
    _filter_interdiffs: bool

    def __init__(
        self,
        differ: Differ,
        diff: Optional[bytes] = None,
        interdiff: Optional[bytes] = None,
        request: Optional[HttpRequest] = None,
        *,
        filediff_orig_lines: (Sequence[str] | None) = None,
        interfilediff_orig_lines: (Sequence[str] | None) = None,
        diff_settings: Optional[DiffSettings] = None,
        **kwargs,
    ) -> None:
        """Initialize the opcode generator.

        Version Changed:
            8.0:
            Added the ``filediff_orig_lines`` and
            ``interfilediff_orig_lines`` parameters, which are now needed for
            interdiff filtering.

        Version Changed:
            7.0.4:
            Added the ``diff_settings`` parameter.

        Version Changed:
            3.0.18:
            Added the ``request`` and ``**kwargs`` parameters.

        Args:
            differ (reviewboard.diffviewer.differ.Differ):
                The differ being used to generate the diff.

            diff (bytes, optional):
                The raw contents for the diff.

            interdiff (bytes, optional):
                The raw contents for the diff on the other end of an
                interdiff range, if generating an interdiff.

            request (django.http.HttpRequest):
                The HTTP request from the client.

            filediff_orig_lines (list of str, optional):
                The normalized lines of the original content of the filediff.

                This is required for interdiff filtering.

                Version Added:
                    8.0

            interfilediff_orig_lines (list of str, optional):
                The normalized lines of the original content of the
                interfilediff.

                This is required for interdiff filtering.

                Version Added:
                    8.0

            diff_settings (reviewboard.diffviewer.settings.DiffSettings,
                           optional):
                The diff settings object.

                Version Added:
                    7.0.4

            **kwargs (dict):
                Additional keyword arguments, for future expansion.
        """
        if diff_settings is None:
            diff_settings = DiffSettings.create(request=request)

        assert diff_settings.tab_size

        filter_interdiffs = bool(
            diff_settings.interdiff_filtering and
            diff and
            interdiff
        )

        if (filter_interdiffs and
            (filediff_orig_lines is None or interfilediff_orig_lines is None)):
            # Interdiff filtering is requested, but new arguments aren't
            # provided. Warn about this and disable interdiff filtering.
            #
            # When we remove this in Review Board 10, raise an exception if
            # these aren't provided.
            class_name = type(self).__name__
            RemovedInReviewBoard10_0Warning.warn(
                f'filediff_orig_lines and interfilediff_orig_lines must be '
                f'passed to {class_name} in order to enable interdiff '
                f'filtering. This will be required in Review Board 10.'
            )
            filter_interdiffs = False

        self.differ = differ
        self.diff = diff
        self.diff_settings = diff_settings
        self.interdiff = interdiff
        self.request = request
        self.filediff_orig_lines = filediff_orig_lines
        self.interfilediff_orig_lines = interfilediff_orig_lines
        self._filter_interdiffs = filter_interdiffs

    def __iter__(self):
        """Returns opcodes from the differ with extra metadata.

        This is a wrapper around a differ's get_opcodes function, which returns
        extra metadata along with each range. That metadata includes
        information on moved blocks of code and whitespace-only lines.

        This returns a list of opcodes as tuples in the form of
        (tag, i1, i2, j1, j2, meta).
        """
        self.groups = []
        self.removes = {}
        self.inserts = []

        # Run the opcodes through the chain.
        opcodes = self.differ.get_opcodes()
        opcodes = self._apply_processors(opcodes)
        opcodes = self._generate_opcode_meta(opcodes)
        opcodes = self._apply_meta_processors(opcodes)

        self._group_opcodes(opcodes)
        self._compute_moves()

        for opcodes in self.groups:
            yield opcodes

    def _apply_processors(self, opcodes):
        """Apply any diff processors to the generated list of opcodes.

        If generating an interdiff, this will apply a filter to remove any
        unmodified lines.

        Args:
            opcodes (list of tuple):
                The list of generated diff opcodes to process.

        Yields:
            tuple:
            A processed opcode.
        """
        if self._filter_interdiffs:
            filediff_orig_lines = self.filediff_orig_lines
            interfilediff_orig_lines = self.interfilediff_orig_lines

            assert filediff_orig_lines is not None
            assert interfilediff_orig_lines is not None

            # Filter out any lines unrelated to these changes from the
            # interdiff. This will get rid of any merge information.
            #
            # Use the newer differ-based approach, which generates ranges
            # consistent with Review Board's own diff algorithm.
            opcodes = filter_interdiff_opcodes(
                opcodes=opcodes,
                differ=self.differ,
                filediff_orig_lines=filediff_orig_lines,
                interfilediff_orig_lines=interfilediff_orig_lines,
                request=self.request,
            )

        yield from opcodes

    def _generate_opcode_meta(self, opcodes):
        for tag, i1, i2, j1, j2 in opcodes:
            meta = {
                # True if this chunk is only whitespace.
                'whitespace_chunk': False,

                # List of tuples (i, j), with whitespace changes.
                'whitespace_lines': [],
            }

            if tag == 'replace':
                # replace groups are good for whitespace only changes.
                assert (i2 - i1) == (j2 - j1)

                for i, j in zip(range(i1, i2), range(j1, j2)):
                    if (self.WHITESPACE_RE.sub('', self.differ.a[i]) ==
                            self.WHITESPACE_RE.sub('', self.differ.b[j])):
                        # Both original lines are equal when removing all
                        # whitespace, so include their original line number in
                        # the meta dict.
                        meta['whitespace_lines'].append((i + 1, j + 1))

                # If all lines are considered to have only whitespace change,
                # the whole chunk is considered a whitespace-only chunk.
                if len(meta['whitespace_lines']) == (i2 - i1):
                    meta['whitespace_chunk'] = True
            elif tag in ('equal', 'filtered-equal'):
                for group in self._compute_chunk_indentation(i1, i2, j1, j2):
                    ii1, ii2, ij1, ij2, indentation_changes = group

                    if indentation_changes:
                        new_meta = dict({
                            'indentation_changes': indentation_changes,
                        }, **meta)

                        yield 'equal', ii1, ii2, ij1, ij2, new_meta
                    else:
                        new_meta = meta

                        yield tag, ii1, ii2, ij1, ij2, new_meta

                continue

            yield tag, i1, i2, j1, j2, meta

    def _apply_meta_processors(self, opcodes):
        if self._filter_interdiffs:
            # When filtering out opcodes, we may have converted chunks into
            # "filtered-equal" chunks. This allowed us to skip any additional
            # processing, particularly the indentation highlighting. It's
            # now time to turn those back into "equal" chunks.
            opcodes = post_process_filtered_equals(opcodes)

        yield from opcodes

    def _group_opcodes(self, opcodes):
        for group_index, group in enumerate(opcodes):
            self.groups.append(group)

            # Store delete/insert ranges for later lookup. We will be building
            # keys that in most cases will be unique for the particular block
            # of text being inserted/deleted. There is a chance of collision,
            # so we store a list of matching groups under that key.
            #
            # Later, we will loop through the keys and attempt to find insert
            # keys/groups that match remove keys/groups.
            tag = group[0]

            if tag in ('delete', 'replace'):
                i1 = group[1]
                i2 = group[2]

                for i in range(i1, i2):
                    line = self.differ.a[i].strip()

                    if line:
                        self.removes.setdefault(line, []).append(
                            (i, group, group_index))

            if tag in ('insert', 'replace'):
                self.inserts.append(group)

    def _compute_chunk_indentation(self, i1, i2, j1, j2):
        """Generate sequential groups of lines with indentation changes.

        This will group together sequential lines that have all had
        indentation changes, and all sequential lines that have not,
        yielding each group. Each group will have a dictionary containing
        the indentation changes for each line in the group.

        Args:
            i1 (int):
                The 0-based index to start processing on the original side.

            i2 (int):
                The 0-based start of the next range on the original side.
                Lines will be processed up to, but not including, this index.

            j1 (int):
                The 0-based index to start processing on the modified side.

            j2 (int):
                The 0-based start of the next range on the modified side.
                Lines will be processed up to, but not including, this index.

        Yields:
            tuple:
            A 5-tuple containing information on a batch of lines with or
            without indentation changes. This includes:

            1. The 0-based start index on the original side.
            2. The 0-based start of the next range to process on the original
               side.
            3. The 0-based start index on the modified side.
            4. The 0-based start of the next range to process on the modified
               side.
            5. A dictionary mapping keys in the form of
               :samp:`"{orig_linenum}-{modified_linenum}"` (1-based) to
               line indentation result tuples (see
               :py:meth:`_compute_line_indentation`).
        """
        # We'll be going through all the opcodes in this equals chunk and
        # grouping with adjacent opcodes based on whether they have
        # indentation changes or not. This allows us to keep the lines with
        # indentation changes from being collapsed in the diff viewer.
        indentation_changes = {}
        prev_has_indent = False
        prev_start_i = i1
        prev_start_j = j1

        a = self.differ.a
        b = self.differ.b

        for i, j in zip(range(i1, i2), range(j1, j2)):
            old_line = a[i]
            new_line = b[j]
            new_indentation_changes = {}

            indent_info = self._compute_line_indentation(old_line, new_line)
            has_indent = indent_info is not None

            if has_indent:
                key = '%d-%d' % (i + 1, j + 1)
                new_indentation_changes[key] = indent_info

            if has_indent != prev_has_indent:
                if prev_start_i != i or prev_start_j != j:
                    # Yield the previous group.
                    yield prev_start_i, i, prev_start_j, j, indentation_changes

                # We have a new group. Set it up, starting with the current
                # calculated state.
                prev_start_i = i
                prev_start_j = j
                prev_has_indent = has_indent
                indentation_changes = new_indentation_changes
            elif has_indent:
                indentation_changes.update(new_indentation_changes)

        # Yield the last group, if we haven't already yielded it.
        if prev_start_i != i2 or prev_start_j != j2:
            yield prev_start_i, i2, prev_start_j, j2, indentation_changes

    def _compute_line_indentation(self, old_line, new_line):
        """Compute the indentation of a line.

        This will determine whether the indentation has changed in a line of
        otherwise "equal" lines.

        Version Changed:
            4.0.7:
            Fixed to avoid indentation calculation for non-equal
            "filtered-equal" lines.

        Args:
            old_line (unicode):
                The old line content.

            new_line (unicode):
                The new line content.

        Returns:
            tuple:
            A 3-tuple if indentation changes were found. This contains:

            1. Whether the content was indented (``True``) or unindented
               (``False``).
            2. How many characters of indentation were added (if indenting)
               or removed (if unindenting).
            3. The difference in indentation levels (between the two lines).

            If no indentation took place, or indentation logic is not
            appropriate for these lines, this will be ``None`` instead.
        """
        if old_line == new_line:
            return None

        old_line_stripped = old_line.lstrip()
        new_line_stripped = new_line.lstrip()

        if old_line_stripped != new_line_stripped:
            # These may be an insert or delete, but they're not equal. We
            # don't want to compute indentation. This is probably in a
            # "filtered-equal".
            return None

        # These are fake-equal. They really have some indentation changes.
        # We want to mark those up.
        #
        # Our goal for this function from here on out is to figure out whether
        # the new line has increased or decreased its indentation, and then
        # to determine how much that has increased or decreased by.
        #
        # Since we may be dealing with the additional or removal of tabs,
        # we have some challenges here. We need to expand those tabs in
        # order to determine if the new line is indented further or not,
        # and then we need to figure out how much of the leading whitespace
        # on either side represents new indentation levels.
        #
        # We do this by chopping off all leading whitespace and expanding
        # any tabs, and then figuring out the total line lengths. That gives
        # us a basis for comparison to determine whether we've indented
        # or unindented.
        #
        # We can then later figure out exactly which indentation characters
        # were added or removed, and then store that information.
        old_line_indent_len = len(old_line) - len(old_line_stripped)
        new_line_indent_len = len(new_line) - len(new_line_stripped)
        old_line_indent = old_line[:old_line_indent_len]
        new_line_indent = new_line[:new_line_indent_len]

        tab_size = self.diff_settings.tab_size
        assert tab_size

        norm_old_line_indent = old_line_indent.expandtabs(tab_size)
        norm_new_line_indent = new_line_indent.expandtabs(tab_size)
        norm_old_line_indent_len = len(norm_old_line_indent)
        norm_new_line_indent_len = len(norm_new_line_indent)
        norm_old_line_len = (norm_old_line_indent_len +
                             len(old_line_stripped))
        norm_new_line_len = (norm_new_line_indent_len +
                             len(new_line_stripped))
        line_len_diff = norm_new_line_len - norm_old_line_len

        if line_len_diff == 0:
            return None

        # We know that a spacing change did take place. We need to figure
        # out now how many characters of indentation were actually
        # added or removed.
        is_indent = (line_len_diff > 0)

        if is_indent:
            raw_indent_len = new_line_indent_len
        else:
            raw_indent_len = old_line_indent_len

        # Figure out how many characters of indentation were in common
        # at the end of the strings. We'll want to exclude these
        # characters when showing indentation changes.
        #
        # This is the area after any new indentation. If the indentation
        # style changed (such as going from tabs to spaces), then nothing
        # will be in common.
        #
        # We figure out the common trailing indentation by reversing both
        # strings and then finding the common prefix. We only care about
        # the length, so we can throw the string away.
        #
        # It may seem odd that we're using os.path.commonprefix, but this
        # isn't really limited to paths. Certainly not in our case. It's
        # worth not re-implementing that logic.
        raw_indent_len -= len(os.path.commonprefix([
            old_line_indent[::-1],
            new_line_indent[::-1],
        ]))

        return (is_indent,
                raw_indent_len,
                abs(norm_old_line_indent_len - norm_new_line_indent_len))

    def _compute_moves(self):
        # We now need to figure out all the moved locations.
        #
        # At this point, we know all the inserted groups, and all the
        # individually deleted lines. We'll be going through and finding
        # consecutive groups of matching inserts/deletes that represent a
        # move block.
        #
        # The algorithm will be documented as we go in the code.
        #
        # We start by looping through all the inserted groups.
        r_move_indexes_used = set()

        for insert in self.inserts:
            self._compute_move_for_insert(r_move_indexes_used, *insert)

    def _compute_move_for_insert(self, r_move_indexes_used, itag, ii1, ii2,
                                 ij1, ij2, imeta):
        """Compute move information for a given insert-like chunk.

        Args:
            r_move_indexes_used (set):
                All remove indexes that have already been included in a move
                range.

            itag (unicode):
                The chunk tag for the insert (``insert`` or ``replace``).

            ii1 (int):
                The 0-based start of the chunk on the original side.

            ii2 (int):
                The 0-based start of the next chunk on the original side.

            ij1 (int):
                The 0-based start of the chunk on the modification side.

            ij2 (int):
                The 0-based start of the next chunk on the modification side.

            imeta (dict):
                The metadata for the chunk for the modification, where the move
                ranges may be stored.
        """
        # Store some state on the range we'll be working with inside this
        # insert group.

        # The current location inside the insert group (from ij1 through ij2).
        i_move_cur = ij1

        # The current range of consecutive lines that we'll use for a move.
        # Each line in this range has a corresponding consecutive delete line.
        i_move_range = MoveRange(i_move_cur, i_move_cur)

        # The deleted move ranges. The key is a string in the form of
        # "{i1}-{i2}-{j1}-{j2}", with those positions taken from the remove
        # group for the line. The value is an instance of MoveRange. The values
        # in MoveRange are used to quickly locate deleted lines we've found
        # that match the inserted lines, so we can assemble ranges later.
        r_move_ranges = {}  # key -> (start, end, group)

        move_key = None
        is_replace = (itag == 'replace')

        # Loop through every location from ij1 through ij2 - 1 until we've
        # reached the end.
        while i_move_cur < ij2:
            try:
                iline = self.differ.b[i_move_cur].strip()
            except IndexError:
                iline = None

            updated_range = False

            if iline and iline in self.removes:
                # The inserted line at this location has a corresponding
                # removed line.
                #
                # If there's already some information on removed line ranges
                # for this particular move block we're processing then we'll
                # update the range.
                #
                # The way we do that is to find each removed line that matches
                # this inserted line, and for each of those find out if there's
                # an existing move range that the found removed line
                # immediately follows. If there is, we update the existing
                # range.
                #
                # If there isn't any move information for this line, we'll
                # simply add it to the move ranges.
                for ri, rgroup, rgroup_index in self.removes[iline]:
                    # Ignore any lines that have already been processed as
                    # part of a move, so we don't end up with incorrect blocks
                    # of lines being matched.
                    if ri in r_move_indexes_used:
                        continue

                    r_move_range = r_move_ranges.get(move_key)

                    if not r_move_range or ri != r_move_range.end + 1:
                        # We either didn't have a previous range, or this
                        # group didn't immediately follow it, so we need
                        # to start a new one.
                        move_key = '%s-%s-%s-%s' % rgroup[1:5]
                        r_move_range = r_move_ranges.get(move_key)

                    if r_move_range:
                        # If the remove information for the line is next in
                        # the sequence for this calculated move range...
                        if ri == r_move_range.end + 1:
                            # This is part of the current range, so update
                            # the end of the range to include it.
                            r_move_range.end = ri
                            r_move_range.add_group(rgroup, rgroup_index)
                            updated_range = True
                    else:
                        # Check that this isn't a replace line that's just
                        # "replacing" itself (which would happen if it's just
                        # changing whitespace).
                        if not is_replace or i_move_cur - ij1 != ri - ii1:
                            # We don't have any move ranges yet, or we're done
                            # with the existing range, so it's time to build
                            # one based on any removed lines we find that
                            # match the inserted line.
                            r_move_ranges[move_key] = \
                                MoveRange(ri, ri, [(rgroup, rgroup_index)])
                            updated_range = True

                    if updated_range:
                        # We found a range we were able to update. Don't
                        # attempt any more matches for removed lines.
                        break

                if not updated_range and r_move_ranges:
                    # We didn't find a move range that this line is a part
                    # of, but we do have some existing move ranges stored.
                    #
                    # Given that updated_range is set, we'll be processing
                    # the known move ranges below. We'll actually want to
                    # re-check this line afterward, so that we can start a
                    # new move range after we've finished processing the
                    # current ones.
                    #
                    # To do that, just i_move_cur back by one. That negates
                    # the increment below.
                    i_move_cur -= 1
                    move_key = None
            elif iline == '' and move_key:
                # This is a blank or whitespace-only line, which would not
                # be in the list of removed lines above. We also have been
                # working on a move range.
                #
                # At this point, the plan is to just attach this blank
                # line onto the end of the last range being operated on.
                #
                # This blank line will help tie together adjacent move
                # ranges. If it turns out to be a trailing line, it'll be
                # stripped later in _determine_move_range.
                r_move_range = r_move_ranges.get(move_key)

                if r_move_range:
                    new_end_i = r_move_range.end + 1

                    if (new_end_i < len(self.differ.a) and
                        self.differ.a[new_end_i].strip() == ''):
                        # There was a matching blank line on the other end
                        # of the range, so we should feel more confident about
                        # adding the blank line here.
                        r_move_range.end = new_end_i

                        # It's possible that this blank line is actually an
                        # "equal" line. Though technically it didn't move,
                        # we're trying to create a logical, seamless move
                        # range, so we need to try to find that group and
                        # add it to the list of groups in the range, if it'
                        # not already there.
                        last_group, last_group_index = r_move_range.last_group

                        if new_end_i >= last_group[2]:
                            # This is in the next group, which hasn't been
                            # added yet. So add it.
                            cur_group_index = r_move_range.last_group[1] + 1
                            r_move_range.add_group(
                                self.groups[cur_group_index],
                                cur_group_index)

                        updated_range = True

            i_move_cur += 1

            if not updated_range or i_move_cur == ij2:
                # We've reached the very end of the insert group. See if
                # we have anything that looks like a move.
                if r_move_ranges:
                    r_move_range = self._find_longest_move_range(r_move_ranges)

                    # If we have a move range, see if it's one we want to
                    # include or filter out. Some moves are not impressive
                    # enough to display. For example, a small portion of a
                    # comment, or whitespace-only changes.
                    r_move_range = self._determine_move_range(r_move_range)

                    if r_move_range:
                        # Rebuild the insert and remove ranges based on where
                        # we are now and which range we won.
                        #
                        # The new ranges will be actual lists of positions,
                        # rather than a beginning and end. These will be
                        # provided to the renderer.
                        #
                        # The ranges expected by the renderers are 1-based,
                        # whereas our calculations for this algorithm are
                        # 0-based, so we add 1 to the numbers.
                        #
                        # The upper boundaries passed to the range() function
                        # must actually be one higher than the value we want.
                        # So, for r_move_range, we actually increment by 2.  We
                        # only increment i_move_cur by one, because i_move_cur
                        # already factored in the + 1 by being at the end of
                        # the while loop.
                        i_range = range(i_move_range.start + 1,
                                        i_move_cur + 1)
                        r_range = range(r_move_range.start + 1,
                                        r_move_range.end + 2)

                        moved_to_ranges = dict(zip(r_range, i_range))

                        for group, group_index in r_move_range.groups:
                            rmeta = group[-1]
                            rmeta.setdefault('moved-to', {}).update(
                                moved_to_ranges)

                        imeta.setdefault('moved-from', {}).update(
                            dict(zip(i_range, r_range)))

                        # Record each of the positions in the removed range
                        # as used, so that they're not factored in again when
                        # determining possible ranges for future moves.
                        #
                        # We'll use the r_range above, but normalize back to
                        # 0-based indexes.
                        r_move_indexes_used.update(r - 1 for r in r_range)

                # Reset the state for the next range.
                move_key = None
                i_move_range = MoveRange(i_move_cur, i_move_cur)
                r_move_ranges = {}

    def _find_longest_move_range(self, r_move_ranges):
        # Go through every range of lines we've found and find the longest.
        #
        # The longest move range wins. If we find two ranges that are equal,
        # though, we'll ignore both. The idea is that if we have two identical
        # moves, then it's probably common enough code that we don't want to
        # show the move. An example might be some standard part of a comment
        # block, with no real changes in content.
        #
        # Note that with the current approach, finding duplicate moves doesn't
        # cause us to reset the winning range to the second-highest identical
        # match. We may want to do that down the road, but it means additional
        # state, and this is hopefully uncommon enough to not be a real
        # problem.
        r_move_range = None

        for iter_move_range in r_move_ranges.values():
            if not r_move_range:
                r_move_range = iter_move_range
            else:
                len1 = r_move_range.end - r_move_range.start
                len2 = iter_move_range.end - iter_move_range.start

                if len1 < len2:
                    r_move_range = iter_move_range
                elif len1 == len2:
                    # If there are two that are the same, it may be common
                    # code that we don't want to see moves for. Comments,
                    # for example.
                    r_move_range = None

        return r_move_range

    def _determine_move_range(self, r_move_range):
        """Determines if a move range is valid and should be included.

        This performs some tests to try to eliminate trivial changes that
        shouldn't have moves associated.

        Specifically, a move range is valid if it has at least one line
        with alpha-numeric characters and is at least 4 characters long when
        stripped.

        If the move range is valid, any trailing whitespace-only lines will
        be stripped, ensuring it covers only a valid range of content.
        """
        if not r_move_range:
            return None

        end_i = r_move_range.end
        lines = self.differ.a[r_move_range.start:end_i + 1]
        new_end_i = None
        valid = False

        for i, line in enumerate(reversed(lines)):
            line = line.strip()

            if line:
                if len(line) >= 4 and self.ALPHANUM_RE.search(line):
                    valid = True

                if new_end_i is None or valid:
                    new_end_i = end_i - i

                if valid:
                    break

        # Accept this if there's more than one line or if the first
        # line is long enough, in order to filter out small bits of garbage.
        valid = (
            valid and
            (new_end_i - r_move_range.start + 1 >=
             self.MOVE_PREFERRED_MIN_LINES or
             len(self.differ.a[r_move_range.start].strip()) >=
             self.MOVE_MIN_LINE_LENGTH))

        if not valid:
            return None

        assert new_end_i is not None

        return MoveRange(r_move_range.start, new_end_i, r_move_range.groups)


_generator = DiffOpcodeGenerator


def get_diff_opcode_generator_class() -> type[DiffOpcodeGenerator]:
    """Return the DiffOpcodeGenerator class used for generating opcodes.

    Returns:
        type:
        The opcode generator class.
    """
    return _generator


def set_diff_opcode_generator_class(
    renderer: type[DiffOpcodeGenerator],
) -> None:
    """Set the DiffOpcodeGenerator class used for generating opcodes.

    Args:
        renderer (type):
            The opcode generator class.
    """
    assert renderer

    globals()['_generator'] = renderer


def get_diff_opcode_generator(*args, **kwargs) -> DiffOpcodeGenerator:
    """Return a DiffOpcodeGenerator instance used for generating opcodes.

    Args:
        *args (tuple):
            Positional arguments to pass to the opcode generator.

        **kwargs (dict):
            Keyword arguments to pass to the opcode generator.

    Returns:
        DiffOpcodeGenerator:
        The new opcode generator.
    """
    return _generator(*args, **kwargs)



"""Unit tests for WebAPITokenManager.

Version Added:
    5.0.5
"""

import datetime
from typing import Optional

import kgb
from django.contrib.auth.models import User
from django.utils import timezone
from djblets.secrets.token_generators.vendor_checksum import \
    VendorChecksumTokenGenerator

from reviewboard.testing.testcase import TestCase
from reviewboard.webapi.models import WebAPIToken


class WebAPITokenManagerTests(kgb.SpyAgency, TestCase):
    """Unit tests for WebAPITokenManager.

    Version Added:
        5.0.5
    """

    #: The token info for created tokens.
    token_info = {'token_type': 'rbp'}

    def setUp(self) -> None:
        """Set up the test case."""
        super().setUp()

        self.token_generator_id = \
            VendorChecksumTokenGenerator.token_generator_id
        self.user = User.objects.create(username='test-user')

        set_time = timezone.make_aware(datetime.datetime(2023, 1, 2, 3))
        self.spy_on(timezone.now, op=kgb.SpyOpReturn(set_time))

    def test_get_or_create_client_token_default(self) -> None:
        """Testing WebAPITokenManager.get_or_create_client_token creates
        a token with default arguments
        """
        settings = {
            'client_token_expiration': 2,
        }

        with self.siteconfig_settings(settings):
            client_token, created = \
                WebAPIToken.objects.get_or_create_client_token(
                    client_name='Test',
                    user=self.user)

            self.assertTrue(created)
            self._assert_client_token_equals(
                token=client_token,
                client_name='Test',
                expires=(timezone.now() + datetime.timedelta(days=2)),
                note='API token automatically created for Test.',
                user=self.user)

    def test_get_or_create_client_token_custom_expires(self) -> None:
        """Testing WebAPITokenManager.get_or_create_client_token creates
        a token with a custom expiration date
        """
        client_token, created = WebAPIToken.objects.get_or_create_client_token(
            client_name='Test',
            expires=timezone.now() + datetime.timedelta(days=10),
            user=self.user)

        self.assertTrue(created)
        self._assert_client_token_equals(
            token=client_token,
            client_name='Test',
            expires=timezone.now() + datetime.timedelta(days=10),
            note='API token automatically created for Test.',
            user=self.user)

    def test_get_or_create_client_token_with_create_default_extra_data(
        self,
    ) -> None:
        """Testing WebAPITokenManager.get_or_create_client_token creates a
        token with a custom default extra_data
        """
        user = self.user

        # An existing non-client token. This shouldn't be returned.
        WebAPIToken.objects.generate_token(
            token_generator_id=self.token_generator_id,
            token_info=self.token_info,
            user=user,
        )

        client_token, created = WebAPIToken.objects.get_or_create_client_token(
            client_name='Test',
            user=self.user,
            default_extra_data={
                'mykey': ['myvalue'],
            },
        )

        self.assertTrue(created)
        self.assertEqual(client_token.extra_data, {
            'mykey': ['myvalue'],
            'client_name': 'Test',
        })

    def test_get_or_create_client_token_with_create_default_policy(
        self,
    ) -> None:
        """Testing WebAPITokenManager.get_or_create_client_token creates a
        token with a custom default policy
        """
        user = self.user

        # An existing non-client token. This shouldn't be returned.
        WebAPIToken.objects.generate_token(
            token_generator_id=self.token_generator_id,
            token_info=self.token_info,
            user=user,
        )

        client_token, created = WebAPIToken.objects.get_or_create_client_token(
            client_name='Test',
            user=self.user,
            default_policy={
                'resources': {
                    '*': {
                        'allow': ['DELETE'],
                    },
                },
            },
        )

        self.assertTrue(created)
        self.assertEqual(client_token.policy, {
            'resources': {
                '*': {
                    'allow': ['DELETE'],
                },
            },
        })

    def test_get_or_create_client_token_with_create_local_site(self) -> None:
        """Testing WebAPITokenManager.get_or_create_client_token creates a
        token with a provided Local Site
        """
        user = self.user
        token_info = self.token_info
        token_generator_id = self.token_generator_id

        local_site1 = self.create_local_site(name='local-site-1')
        local_site2 = self.create_local_site(name='local-site-2')

        # An existing client token on the global site.
        WebAPIToken.objects.generate_token(
            token_generator_id=token_generator_id,
            token_info=token_info,
            user=user,
            extra_data={
                'client_name': 'Test',
            },
        )

        # An existing client token on a different Local Site.
        WebAPIToken.objects.generate_token(
            token_generator_id=token_generator_id,
            token_info=token_info,
            user=user,
            extra_data={
                'client_name': 'Test',
            },
            local_site=local_site1,
        )

        client_token, created = WebAPIToken.objects.get_or_create_client_token(
            client_name='Test',
            user=user,
            local_site=local_site2,
        )

        self.assertTrue(created)
        self.assertEqual(client_token.local_site, local_site2)

    def test_get_or_create_client_token_custom_expires_none(self) -> None:
        """Testing WebAPITokenManager.get_or_create_client_token creates
        a token with no expiration date when expires=None
        """
        client_token, created = WebAPIToken.objects.get_or_create_client_token(
            client_name='Test',
            expires=None,
            user=self.user)

        self.assertTrue(created)
        self._assert_client_token_equals(
            token=client_token,
            client_name='Test',
            expires=None,
            note='API token automatically created for Test.',
            user=self.user)

    def test_get_or_create_client_token_with_existing_extra_data(self) -> None:
        """Testing WebAPITokenManager.get_or_create_client_token returns
        an existing client token when multiple fields exist in the extra data
        """
        token = WebAPIToken.objects.generate_token(
            extra_data={
                'field_1': 'Some field',
                'client_name': 'Test',
                'field_2': 25,
            },
            token_generator_id=self.token_generator_id,
            token_info=self.token_info,
            user=self.user)

        # An existing non-client token. This shouldn't be returned.
        WebAPIToken.objects.generate_token(
            token_generator_id=self.token_generator_id,
            token_info=self.token_info,
            user=self.user)

        client_token, created = WebAPIToken.objects.get_or_create_client_token(
            client_name='Test',
            user=self.user)

        self.assertFalse(created)
        self.assertEqual(token, client_token)

    def test_get_or_create_client_token_with_existing(self) -> None:
        """Testing WebAPITokenManager.get_or_create_client_token returns
        an existing client token
        """
        token = WebAPIToken.objects.generate_token(
            extra_data={
                'client_name': 'Test',
            },
            token_generator_id=self.token_generator_id,
            token_info=self.token_info,
            user=self.user)

        # An existing non-client token. This shouldn't be returned.
        WebAPIToken.objects.generate_token(
            token_generator_id=self.token_generator_id,
            token_info=self.token_info,
            user=self.user)

        client_token, created = WebAPIToken.objects.get_or_create_client_token(
            client_name='Test',
            user=self.user)

        self.assertFalse(created)
        self.assertEqual(token, client_token)

    def test_get_or_create_client_token_with_existing_ignore_deprecated(
        self,
    ) -> None:
        """Testing WebAPITokenManager.get_or_create_client_token returns
        an existing client token and ignore_deprecated=True
        """
        user = self.user

        # A deprecated API token.
        token = WebAPIToken.objects.generate_token(
            extra_data={
                'client_name': 'Test',
            },
            token_generator_id='legacy_sha1',
            token_info={
                'attempt': 1,
                'user': user,
            },
            user=user,
        )

        client_token, created = WebAPIToken.objects.get_or_create_client_token(
            client_name='Test',
            user=user,
            ignore_deprecated=True,
        )

        self.assertTrue(created)
        self.assertNotEqual(token, client_token)

    def test_get_or_create_client_token_with_expired(self) -> None:
        """Testing WebAPITokenManager.get_or_create_client_token creates a new
        client token when the existing client token is expired
        """
        expired_token = WebAPIToken.objects.generate_token(
            expires=timezone.now() - datetime.timedelta(days=2),
            extra_data={
                'client_name': 'Test',
            },
            token_generator_id=self.token_generator_id,
            token_info=self.token_info,
            user=self.user)

        # An existing non-client token. This shouldn't be returned.
        WebAPIToken.objects.generate_token(
            expires=timezone.now() + datetime.timedelta(days=50),
            token_generator_id=self.token_generator_id,
            token_info=self.token_info,
            user=self.user)

        client_token, created = WebAPIToken.objects.get_or_create_client_token(
            client_name='Test',
            user=self.user)

        self.assertTrue(created)
        self.assertNotEqual(expired_token, client_token)

    def test_get_or_create_client_token_with_expires_soon(self) -> None:
        """Testing WebAPITokenManager.get_or_create_client_token creates a new
        client token when the existing client token is expiring soon (under
        the minimum validity period)
        """
        user = self.user
        token_info = self.token_info
        token_generator_id = self.token_generator_id

        expired_token = WebAPIToken.objects.generate_token(
            expires=timezone.now() + datetime.timedelta(seconds=4 * 60 + 59),
            extra_data={
                'client_name': 'Test',
            },
            token_generator_id=token_generator_id,
            token_info=token_info,
            user=user,
        )

        # An existing non-client token. This shouldn't be returned.
        WebAPIToken.objects.generate_token(
            expires=timezone.now() + datetime.timedelta(days=50),
            token_generator_id=token_generator_id,
            token_info=token_info,
            user=user)

        client_token, created = WebAPIToken.objects.get_or_create_client_token(
            client_name='Test',
            user=user,
        )

        self.assertTrue(created)
        self.assertNotEqual(expired_token, client_token)

    def test_get_or_create_client_token_with_invalid(self) -> None:
        """Testing WebAPITokenManager.get_or_create_client_token creates a new
        client token when the existing client token is invalid
        """
        expired_token = WebAPIToken.objects.generate_token(
            expires=None,
            extra_data={
                'client_name': 'Test',
            },
            token_generator_id=self.token_generator_id,
            token_info=self.token_info,
            user=self.user,
            valid=False)

        # An existing non-client token. This shouldn't be returned.
        WebAPIToken.objects.generate_token(
            expires=timezone.now() + datetime.timedelta(days=50),
            token_generator_id=self.token_generator_id,
            token_info=self.token_info,
            user=self.user)

        client_token, created = WebAPIToken.objects.get_or_create_client_token(
            client_name='Test',
            user=self.user)

        self.assertTrue(created)
        self.assertNotEqual(expired_token, client_token)

    def test_get_or_create_client_token_sorted_by_expires(self) -> None:
        """Testing WebAPITokenManager.get_or_create_client_token returns the
        client token with the furthest expiration date when multiple ones exist
        """
        WebAPIToken.objects.generate_token(
            expires=timezone.now() - datetime.timedelta(days=2),
            extra_data={
                'client_name': 'Test',
            },
            token_generator_id=self.token_generator_id,
            token_info=self.token_info,
            user=self.user)

        # This non-client token with the furthest expiration shouldn't
        # be returned.
        WebAPIToken.objects.generate_token(
            expires=timezone.now() + datetime.timedelta(days=50),
            token_generator_id=self.token_generator_id,
            token_info=self.token_info,
            user=self.user)

        token_furthest = WebAPIToken.objects.generate_token(
            expires=timezone.now() + datetime.timedelta(days=30),
            extra_data={
                'client_name': 'Test',
            },
            token_generator_id=self.token_generator_id,
            token_info=self.token_info,
            user=self.user)

        client_token, created = WebAPIToken.objects.get_or_create_client_token(
            client_name='Test',
            user=self.user)

        self.assertFalse(created)
        self.assertEqual(token_furthest, client_token)

    def test_get_or_create_client_token_sorted_by_expires_none(self) -> None:
        """Testing WebAPITokenManager.get_or_create_client_token returns the
        client token with no expiration date when multiple ones exist
        """
        token_furthest = WebAPIToken.objects.generate_token(
            expires=None,
            extra_data={
                'client_name': 'Test',
            },
            token_generator_id=self.token_generator_id,
            token_info=self.token_info,
            user=self.user)

        WebAPIToken.objects.generate_token(
            expires=timezone.now() + datetime.timedelta(days=365),
            extra_data={
                'client_name': 'Test',
            },
            token_generator_id=self.token_generator_id,
            token_info=self.token_info,
            user=self.user)

        client_token, created = WebAPIToken.objects.get_or_create_client_token(
            client_name='Test',
            user=self.user)

        self.assertFalse(created)
        self.assertEqual(token_furthest, client_token)

    def _assert_client_token_equals(
        self,
        token: WebAPIToken,
        client_name: str,
        expires: Optional[datetime.datetime],
        note: str,
        user: User,
    ) -> None:
        """Assert that the client token matches the given values.

        Args:
            token (reviewboard.webapi.models.WebAPIToken):
                The token.

            client_name (str):
                The client for the token.

            expires (datetime.datetime):
                The expiration date for the token.

            note (str):
                The note for the token.

            user (django.contrib.auth.models.User):
                The user for the token.

        Raises:
            AssertionError:
                The token is not invalid.
        """
        self.assertEqual(token.user, user)
        self.assertEqual(token.extra_data['client_name'], client_name)
        self.assertEqual(token.note, note)
        self.assertEqual(token.expires, expires)



"""ClearCase SCM provider."""

import logging
import os
import re
import subprocess
import sys
import tempfile

from django.conf import settings
from django.utils.encoding import force_str
from django.utils.translation import gettext_lazy as _

from reviewboard.diffviewer.parser import DiffParser
from reviewboard.scmtools.core import SCMTool, HEAD, PRE_CREATION
from reviewboard.scmtools.errors import SCMError, FileNotFoundError

# This specific import is necessary to handle the paths for
# cygwin enabled machines.
if sys.platform.startswith(('win', 'cygwin')):
    import ntpath as cpath
else:
    import posixpath as cpath


logger = logging.getLogger(__name__)


_cleartool = None


def get_cleartool():
    """Return the cleartool binary/path name.

    This allows the user to configure a custom path to cleartool, or use a
    wrapper, by setting CC_CTEXEC in the settings_local.py file.

    Returns:
        unicode:
        The name or path of cleartool to use.
    """
    global _cleartool

    if _cleartool is None:
        _cleartool = getattr(settings, 'CC_CTEXEC', None)

        if not _cleartool:
            _cleartool = 'cleartool'

        logger.debug('Using cleartool %s', _cleartool)

    return _cleartool


class ClearCaseTool(SCMTool):
    """ClearCase SCM provider."""

    scmtool_id = 'clearcase'
    name = 'ClearCase'
    field_help_text = {
        'path': _('The absolute path to the VOB.'),
    }
    dependencies = {
        'executables': [get_cleartool()],
    }

    # This regular expression can extract from extended_path pure system path.
    # It is construct from two main parts. The first match is everything from
    # beginning of line to the first occurrence of /. The second match is
    # parts between /main and numbers (file version). This patch assumes each
    # branch present in extended_path was derived from /main and there is no
    # file or directory called "main" in path.
    UNEXTENDED = re.compile(r'^(.+?)/|/?(.+?)/main/?.*?/([0-9]+|CHECKEDOUT)')

    # Currently, snapshot and dynamic views are supported. Automatic views and
    # webviews will be reported as VIEW_UNKNOWN.
    VIEW_SNAPSHOT, VIEW_DYNAMIC, VIEW_UNKNOWN = range(3)

    def __init__(self, repository):
        """Initialize the tool.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The associated repository object.
        """
        self.repopath = repository.path

        SCMTool.__init__(self, repository)

        self.viewtype = self._get_view_type(self.repopath)

        if self.viewtype == self.VIEW_SNAPSHOT:
            self.client = ClearCaseSnapshotViewClient(self.repopath)
        elif self.viewtype == self.VIEW_DYNAMIC:
            self.client = ClearCaseDynamicViewClient(self.repopath)
        else:
            raise SCMError('Unsupported view type.')

    @staticmethod
    def run_cleartool(cmdline, cwd=None, ignore_errors=False,
                      results_unicode=True):
        """Run cleartool with the given command line.

        Args:
            cmdline (list of unicode):
                The cleartool command-line to execute.

            cwd (unicode, optional):
                The working directory to use for the subprocess.

            ignore_errors (bool, optional):
                Whether to ignore error return codes.

            results_unicode (bool, optional):
                Whether to return unicode or bytes.

        Returns:
            bytes or unicode:
            The output from the command.

        Raises:
            reviewboard.scmtools.errors.SCMError:
                The cleartool execution returned an error code.
        """
        popen_kwargs = {}

        if results_unicode:
            # Popen before Python 3.6 doesn't support the ``encoding``
            # parameter, so we have to use ``universal_newlines`` and then
            # decode later.
            if sys.version_info[:2] >= (3, 6):
                popen_kwargs['encoding'] = 'utf-8'
            else:
                popen_kwargs['universal_newlines'] = True

        # On Windows 7+, executing a process that is marked SUBSYSTEM_CONSOLE
        # (such as cleartool) will pop up a console window, even if output is
        # redirected to a pipe. This hot mess prevents that from happening. If
        # Popen gains a better API to do this, we should switch to that when
        # we can. See https://bugs.python.org/issue30082 for details.
        if sys.platform.startswith('win'):
            si = subprocess.STARTUPINFO()
            si.dwFlags |= subprocess.STARTF_USESHOWWINDOW
            si.wShowWindow = subprocess.SW_HIDE
            popen_kwargs['startupinfo'] = si

        cmdline = [get_cleartool()] + cmdline

        logger.debug('Running %s', subprocess.list2cmdline(cmdline))

        p = subprocess.Popen(
            cmdline,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=cwd,
            **popen_kwargs)

        (results, error) = p.communicate()
        failure = p.returncode

        if failure and not ignore_errors:
            raise SCMError(error)

        # We did not specify ``encoding`` to Popen earlier, so decode now.
        if results_unicode and 'encoding' not in popen_kwargs:
            results = force_str(results)

        return results

    def normalize_path_for_display(self, filename, extra_data=None, **kwargs):
        """Normalize a path from a diff for display to the user.

        This will strip away information about the branch, version, and
        repository path, returning an unextended path relative to the view.

        ClearCase paths contain additional information about branch and file
        version preceded by @@. This function removes these parts from the
        ClearCase path to make it more readable. For example this function
        converts the extended path::

            /vobs/comm@@/main/122/network@@/main/55/sntp
            @@/main/4/src@@/main/1/sntp.c@@/main/8

        to the the to regular path::

            /vobs/comm/network/sntp/src/sntp.c

        Args:
            filename (unicode):
                The filename/path to normalize.

            extra_data (dict, optional):
                Extra data stored for the diff this file corresponds to.
                This may be empty or ``None``. Subclasses should not assume the
                presence of anything here.

            **kwargs (dict, unused):
                Additional keyword arguments.

        Returns:
            unicode:
            The resulting filename/path.

        """
        if '@@' not in filename:
            return filename

        # The result of regular expression search is a list of tuples. We must
        # flatten this to a single list. b is first because it frequently
        # occurs in tuples. Before that remove @@ from path.
        unextended_chunks = [
            b or a
            for a, b, foo in self.UNEXTENDED.findall(
                filename.replace('@@', ''))
        ]

        if sys.platform.startswith('win'):
            # Properly handle full (with drive letter) and UNC paths.
            if unextended_chunks[0].endswith(':'):
                unextended_chunks[0] = '%s\\' % unextended_chunks[0]
            elif unextended_chunks[0] == '/' or unextended_chunks[0] == os.sep:
                unextended_chunks[0] = '\\\\'

        unextended_path = cpath.join(*unextended_chunks)

        if not cpath.isabs(unextended_path):
            unextended_path = cpath.join(self.repopath, unextended_path)

        # Calling realpath here removes extra garbage like ``/./`` which can
        # be generated by ClearCase when a vob branch is newly created.
        unextended_path = cpath.realpath(unextended_path)

        return cpath.relpath(unextended_path, self.repopath)

    def get_repository_info(self):
        """Return repository information.

        Returns:
            dict:
            A dictionary containing information for the repository, including
            the VOB tag and UUID.
        """
        vobstag = self._get_vobs_tag(self.repopath)
        return {
            'repopath': self.repopath,
            'uuid': self._get_vobs_uuid(vobstag)
        }

    def _get_view_type(self, repopath):
        """Return the ClearCase view type for the given path.

        Args:
            repopath (unicode):
                The repository path.

        Returns:
            int:
            One of :py:attr:`VIEW_SNAPSHOT`, :py:attr:`VIEW_DYNAMIC`, or
            :py:attr:`VIEW_UNKNOWN`.

        Raises:
            reviewboard.scmtools.errors.SCMError:
                An error occurred when finding the view type.
        """
        result = self.run_cleartool(
            ['lsview', '-full', '-properties', '-cview'],
            cwd=repopath)

        for line in result.splitlines(True):
            splitted = line.split(' ')

            if splitted[0] == 'Properties:':
                if 'snapshot' in splitted:
                    return self.VIEW_SNAPSHOT
                elif 'dynamic' in splitted:
                    return self.VIEW_DYNAMIC

        return self.VIEW_UNKNOWN

    def _get_vobs_tag(self, repopath):
        """Return the VOB tag for the given path.

        Args:
            repopath (unicode):
                The repository path.

        Returns:
            unicode:
            The VOB tag for the repository at the given path.

        Raises:
            reviewboard.scmtools.errors.SCMError:
                An error occurred when finding the VOB tag.
        """
        result = self.run_cleartool(
            ['describe', '-short', 'vob:.'],
            cwd=repopath)

        return result.rstrip()

    def _get_vobs_uuid(self, vobstag):
        """Return the UUID for the given VOB tag.

        Args:
            vobstag (unicode):
                The VOB tag.

        Returns:
            unicode:
            The UUID associated with the given VOB tag.

        Raises:
            reviewboard.scmtools.errors.SCMError:
                An error occurred when finding the UUID.
        """
        result = self.run_cleartool(
            ['lsvob', '-long', vobstag],
            cwd=self.repopath)

        for line in result.splitlines(True):
            if line.startswith('Vob family uuid:'):
                return line.split(' ')[-1].rstrip()

        raise SCMError('Unable to find family uuid for vob: %s' % vobstag)

    def _get_element_kind(self, extended_path):
        """Return the element type of a VOB element.

        Args:
            extended_path (unicode):
                The path of the element, including revision information.

        Returns:
            unicode:
            The element type of the given element.

        Raises:
            reviewboard.scmtools.errors.SCMError:
                An error occurred when finding the element type.
        """
        result = self.run_cleartool(
            ['desc', '-fmt', '%m', extended_path],
            cwd=self.repopath)

        return result.strip()

    def get_file(self, extended_path, revision=HEAD, **kwargs):
        """Return content of file or list content of directory.

        Args:
            extended_path (unicode):
                The path of the element, including revision information.

            revision (reviewboard.scmtools.core.Revision, optional):
                Revision information. This will be either
                :py:data:`~reviewboard.scmtools.core.PRE_CREATION` (new file),
                or :py:data:`~reviewboard.scmtools.core.HEAD` (signifying to
                use the revision information included in ``extended_path``).

            **kwargs (dict, optional):
                Additional unused keyword arguments.

        Returns:
            bytes:
            The contents of the element.

        Raises:
            reviewboard.scmtools.errors.FileNotFoundError:
                The given ``extended_path`` did not match a valid element.

            reviewboard.scmtools.errors.SCMError:
                Another error occurred.
        """
        if not extended_path:
            raise FileNotFoundError(extended_path, revision)

        if revision == PRE_CREATION:
            return ''

        if self.viewtype == self.VIEW_SNAPSHOT:
            # Get the path to (presumably) file element (remove version)
            # The '@@' at the end of file_path is required.
            file_path = extended_path.rsplit('@@', 1)[0] + '@@'
            okind = self._get_element_kind(file_path)

            if okind == 'directory element':
                raise SCMError('Directory elements are unsupported.')
            elif okind == 'file element':
                output = self.client.cat_file(extended_path)
            else:
                raise FileNotFoundError(extended_path)
        else:
            if cpath.isdir(extended_path):
                output = self.client.list_dir(extended_path)
            elif cpath.exists(extended_path):
                output = self.client.cat_file(extended_path)
            else:
                raise FileNotFoundError(extended_path)

        return output

    def parse_diff_revision(self, filename, revision, *args, **kwargs):
        """Parse and return a filename and revision from a diff.

        In the diffs for ClearCase, the revision is actually part of the file
        path. The ``revision_str`` argument contains modification timestamps.

        Args:
            filename (bytes):
                The filename as represented in the diff.

            revision (bytes):
                The revision as represented in the diff.

            *args (tuple, unused):
                Unused positional arguments.

            **kwargs (dict, unused):
                Unused keyword arguments.

        Returns:
            tuple:
            A tuple containing two items:

            1. The normalized filename as a byte string.
            2. The normalized revision as a byte string or a
               :py:class:`~reviewboard.scmtools.core.Revision`.
        """
        assert isinstance(filename, bytes), (
            'filename must be a byte string, not %r' % type(filename))
        assert isinstance(revision, bytes), (
            'revision must be a byte string, not %r' % type(revision))

        if filename.endswith(os.path.join(os.sep, 'main',
                                          '0').encode('utf-8')):
            revision = PRE_CREATION
        elif filename.endswith(b'CHECKEDOUT') or b'@@' not in filename:
            revision = HEAD
        else:
            revision = filename.rsplit(b'@@', 1)[1]

        return filename, revision

    def get_parser(self, data):
        """Return the diff parser for a ClearCase diff.

        Args:
            data (bytes):
                The diff data.

        Returns:
            ClearCaseDiffParser:
            The diff parser.
        """
        return ClearCaseDiffParser(data,
                                   self.repopath,
                                   self._get_vobs_tag(self.repopath))


class ClearCaseDiffParser(DiffParser):
    """Special parser for ClearCase diffs created with RBTools."""

    SPECIAL_REGEX = re.compile(br'^==== (\S+) (\S+) ====$')

    def __init__(self, data, repopath, vobstag):
        """Initialize the parser.

        Args:
            data (bytes):
                The diff data.

            repopath (unicode):
                The path to the repository.

            vobstag (unicode):
                The VOB tag for the repository.
        """
        self.repopath = repopath
        self.vobstag = vobstag
        super(ClearCaseDiffParser, self).__init__(data)

    def parse_diff_header(self, linenum, info):
        """Parse a diff header.

        Paths for the same file may differ from paths in developer view because
        it depends from configspec and this is custom so we translate oids,
        attached by RBTools, to filenames to get paths working well inside
        clearcase view on reviewboard side.

        Args:
            linenum (int):
                The line number to start parsing at.

            info (dict):
                The diff info data to populate.

        Returns:
            int:
            The line number after the end of the diff header.
        """
        # Because ==== oid oid ==== is present after each header
        # parse standard +++ and --- headers at the first place
        linenum = super(ClearCaseDiffParser, self).parse_diff_header(
            linenum, info)

        # Parse for filename.
        m = self.SPECIAL_REGEX.match(self.lines[linenum])

        if m:
            # When using ClearCase in multi-site mode, data replication takes
            # much time, including oid. As said above, oid is used to retrieve
            # filename path independent of developer view.
            # When an oid is not found on server side an exception is thrown
            # and review request submission fails.
            # However at this time origFile and newFile info have already been
            # filled by super.parse_diff_header and contain client side paths,
            # client side paths are enough to start reviewing.
            # So we can safely catch exception and restore client side paths
            # if not found.
            # Note: origFile and newFile attributes are not defined when
            # managing binaries, so init to '' as fallback.
            current_filename = info.get('origFile', '')

            try:
                info['origFile'] = self._oid2filename(m.group(1))
            except Exception:
                logger.debug('oid (%s) not found, get filename from client',
                             m.group(1))
                info['origFile'] = self.client_relpath(current_filename)

            current_filename = info.get('newFile', '')

            try:
                info['newFile'] = self._oid2filename(m.group(2))
            except Exception:
                logger.debug('oid (%s) not found, get filename from client',
                             m.group(2))
                info['newFile'] = self.client_relpath(current_filename)

            linenum += 1

            if (linenum < len(self.lines) and
                (self.lines[linenum].startswith((b'Binary files ',
                                                 b'Files ')))):
                # To consider filenames translated from oids
                # origInfo and newInfo keys must exists.
                # Other files already contain this values field
                # by timestamp from +++/--- diff header.
                info['origInfo'] = ''
                info['newInfo'] = ''

                # Binary files need add origInfo and newInfo manually
                # because they don't have diff's headers (only oids).
                info['binary'] = True
                linenum += 1

        return linenum

    def _oid2filename(self, oid):
        """Convert an oid to a filename.

        Args:
            oid (unicode):
                The given oid.

        Returns:
            unicode:
            The filename of the element relative to the repopath.

        Raises:
            reviewboard.scmtools.errors.SCMError:
                An error occurred while finding the filename.
        """
        result = ClearCaseTool.run_cleartool(
            ['describe', '-fmt', '%En@@%Vn', 'oid:%s' % oid],
            cwd=self.repopath)

        drive = os.path.splitdrive(self.repopath)[0]

        if drive:
            result = os.path.join(drive, result)

        return cpath.relpath(result, self.repopath)

    def client_relpath(self, filename):
        """Normalize a client view path.

        Args:
            filename (unicode):
                A path in a client view.

        Returns:
            unicode:
            The relative path against the vobstag.
        """
        try:
            path, revision = filename.split(b'@@', 1)
        except ValueError:
            path = filename
            revision = None

        relpath = b''
        tries = 0
        vobtag_bytes = self.vobstag.encode('utf-8')
        logger.debug('vobstag: %s, path: %s', self.vobstag, path)

        while tries < 20:
            # An error should be raised if vobstag cannot be reached.
            if path == b'/':
                logger.debug('vobstag not found in path, use client filename')
                return filename

            # Vobstag reach, relpath can be returned.
            if path.endswith(vobtag_bytes):
                break

            path, basename = os.path.split(path)

            # Init relpath with basename.
            if len(relpath) == 0:
                relpath = basename
            else:
                relpath = os.path.join(basename, relpath)

            tries += 1
        else:
            # A correctly formatted diff from RBTools will have the vobtag in
            # the filenames, but if its not there, we don't want to infinitely
            # loop trying to find it.
            logger.debug('Failed to find vobtag prefix in filename %s, '
                         'aborting',
                         filename)

        logger.debug('relpath: %s', relpath)

        if revision:
            relpath = relpath + b'@@' + revision

        return relpath


class ClearCaseDynamicViewClient(object):
    """A client for ClearCase dynamic views."""

    def __init__(self, path):
        """Initialize the client.

        Args:
            path (unicode):
                The path of the view.
        """
        self.path = path

    def cat_file(self, extended_path):
        """Return the contents of a file at a given revision.

        Args:
            extended_path (unicode):
                The file to fetch. This includes revision information.

        Returns:
            bytes:
            The contents of the file.
        """
        # As we are in a dynamic view, we can use the extended pathname to
        # access the file directly.
        with open(extended_path, 'rb') as f:
            return f.read()

    def list_dir(self, extended_path):
        """Return a directory listing of the given path.

        Args:
            extended_path (unicode):
                The path to the directory. This includes revision information.

        Returns:
            unicode:
            The contents of the given directory.
        """
        # As we are in a dynamic view, we can use the extended pathname to
        # access the directory directly.
        return ''.join([
            '%s\n' % s
            for s in sorted(os.listdir(extended_path))
        ])


class ClearCaseSnapshotViewClient(object):
    """A client for ClearCase snapshot views."""

    def __init__(self, path):
        """Initialize the client.

        Args:
            path (unicode):
                The path of the view.
        """
        self.path = path

    def cat_file(self, extended_path):
        """Return the contents of a file at a given revision.

        Args:
            extended_path (unicode):
                The file to fetch. This includes revision information.

        Returns:
            bytes:
            The contents of the file.

        Raises:
            reviewboard.scmtools.errors.FileNotFoundError:
                The given ``extended_path`` did not match a valid element.
        """
        # In a snapshot view, we cannot directly access the file. Use cleartool
        # to pull the desired revision into a temp file.
        temp = tempfile.NamedTemporaryFile()

        # Close and delete the existing file so we can write to it.
        temp.close()

        try:
            ClearCaseTool.run_cleartool(
                ['get', '-to', temp.name, extended_path])
        except SCMError:
            raise FileNotFoundError(extended_path)

        try:
            with open(temp.name, 'rb') as f:
                return f.read()
        except Exception:
            raise FileNotFoundError(extended_path)
        finally:
            try:
                os.unlink(temp.name)
            except Exception:
                pass



"""Model managers for Repository and Tool."""

from __future__ import annotations

import logging
from typing import Any, Optional, Sequence, TYPE_CHECKING, Union

import importlib_metadata
from django.db.models import Manager, Q
from django.db.models.query import QuerySet
from housekeeping.functions import deprecate_non_keyword_only_args

from reviewboard.deprecation import RemovedInReviewBoard90Warning
from reviewboard.site.models import LocalSite

if TYPE_CHECKING:
    from django.contrib.auth.models import AnonymousUser, User

    from reviewboard.scmtools.models import Repository, Tool
    from reviewboard.site.models import AnyOrAllLocalSites


logger = logging.getLogger(__name__)


_TOOL_CACHE: dict[Any, Tool] = {}


class ToolQuerySet(QuerySet):
    """QuerySet for accessing database-registered SCMTools.

    This provides some basic caching capabilities to ensure that common
    lookups of tools don't hit the database any more than necessary.
    """

    def get(self, *args, **kwargs) -> Tool:
        """Return a Tool registration from the database.

        If querying directly by ID, this will return a cached entry, if
        available.

        If the cache is empty, all database registrations will be queried
        immediately and cached.

        And other queries will proceed as normal, uncached.

        Args:
            *args (tuple):
                Positional query arguments.

            **kwargs (dict):
                Keyword query arguments.

        Returns:
            reviewboard.scmtools.models.Tool:
            The queried Tool, if found.

        Raises:
            reviewboard.scmtools.models.Tool.DoesNotExist:
                The queried Tool could not be found.

            reviewboard.scmtools.models.Tool.MultipleObjectsReturned:
                Multiple Tools matching the query were found.
        """
        pk: Any = None

        # This is all pretty awful. We're not meant to reach into these
        # objects. However, we also don't really have another way of finding
        # out what ID was queried.
        #
        # This will be less of a problem when we move away from database-backed
        # Tool registration.
        if len(args) > 0 and isinstance(args[0], Q):
            try:
                query_field, query_value = args[0].children[0]
            except Exception:
                query_field = None
                query_value = None

            if query_field in ('id', 'id__exact', 'pk', 'pk__exact'):
                pk = query_value
        else:
            pk = (kwargs.get('pk') or
                  kwargs.get('pk__exact') or
                  kwargs.get('id') or
                  kwargs.get('id__exact'))

        if pk is None:
            # Something else was queried. Request it from the database as
            # normal.
            return super(ToolQuerySet, self).get(*args, **kwargs)

        if not _TOOL_CACHE:
            # Precompute the cache to reduce lookups.
            for tool in self.model.objects.all():
                _TOOL_CACHE[tool.pk] = tool

        if pk not in _TOOL_CACHE:
            # We'll try to look up the Tool anyway, since it may have been
            # added since. This will also ensure the proper exception is
            # raised if not found.
            _TOOL_CACHE[pk] = super(ToolQuerySet, self).get(*args, **kwargs)

        return _TOOL_CACHE[pk]


class ToolManager(Manager['Tool']):
    """Manages Tool models.

    Any get() operations performed (directly or indirectly through a
    ForeignKey) will go through a cache to attempt to minimize Tool
    lookups.

    The Tool cache is never cleared, but as Tool objects should never
    be modified by hand (they're registered when doing an rb-site upgrade,
    and then the server process must be reloaded), this shouldn't be a
    problem.
    """

    use_for_related_fields = True

    def register_from_entrypoints(self) -> Sequence[Tool]:
        """Register tools from any package-provided Python Entrypoints.

        This will add any new tools that aren't already in the database.

        Returns:
            list of reviewboard.scmtools.models.Tool:
            The list of new tools added to the database.
        """
        registered_tools = set(self.values_list('class_name', flat=True))
        new_tools: list[Tool] = []

        eps = importlib_metadata.entry_points(group='reviewboard.scmtools')

        for entry in eps:
            try:
                scmtool_class = entry.load()
            except Exception as e:
                logger.exception('Unable to load SCMTool %s: %s',
                                 entry, e)
                continue

            class_name = '%s.%s' % (scmtool_class.__module__,
                                    scmtool_class.__name__)

            if class_name not in registered_tools:
                registered_tools.add(class_name)
                name = (scmtool_class.name or
                        scmtool_class.__name__.replace('Tool', ''))

                new_tools.append(self.model(class_name=class_name,
                                            name=name))

        if new_tools:
            self.bulk_create(new_tools)

        return new_tools

    def get_queryset(self) -> ToolQuerySet:
        """Return a QuerySet for Tool models.

        Returns:
            ToolQuerySet:
            The new QuerySet instance.
        """
        return ToolQuerySet(self.model, using=self.db)

    def clear_tool_cache(self) -> None:
        """Clear the internal cache of Tools.

        This is intended for unit tests, and won't be called during production.
        """
        _TOOL_CACHE.clear()


class RepositoryManager(Manager['Repository']):
    """A manager for Repository models."""

    @deprecate_non_keyword_only_args(RemovedInReviewBoard90Warning)
    def accessible(
        self,
        user: Union[AnonymousUser, User],
        *,
        visible_only: bool = True,
        local_site: AnyOrAllLocalSites = None,
        distinct: bool = True,
    ) -> QuerySet[Repository]:
        """Return a queryset for repositories accessible by the given user.

        For superusers, all public and private repositories will be returned.

        For regular users, only repositories that are public or that the user
        is on the access lists for (directly or through a review group) will
        be returned.

        For anonymous users, only public repositories will be returned.

        The returned list is further filtered down based on the
        ``visible_only`` and ``local_site`` parameters.

        Version Changed:
            6.0:
            Removed the ``show_all_local_sites`` argument.

        Version Changed:
            5.0:
            Deprecated ``show_all_local_sites`` and added support for
            setting ``local_site`` to :py:class:`LocalSite.ALL
            <reviewboard.site.models.LocalSite.ALL>`.

        Version Changed:
            3.0.24:
            Added the ``distinct`` parameter.

        Args:
            user (django.contrib.auth.models.User):
                The user that must have access to any returned repositories.

            visible_only (bool, optional):
                Whether only visible repositories should be returned.

            local_site (reviewboard.site.models.LocalSite or
                        reviewboard.site.models.LocalSite.ALL, optional):
                A specific :term:`Local Site` that the repositories must be
                associated with. By default, this will only return
                repositories not part of a site.

                This may be :py:attr:`LocalSite.ALL
                <reviewboard.site.models.LocalSite.ALL>`.

                Version Changed:
                    5.0:
                    Added support for :py:attr:`LocalSite.ALL
                    <reviewboard.site.models.LocalSite.ALL>`.

            distinct (bool, optional):
                Whether to return distinct results.

                Turning this off can increase performance. It's on by default
                for backwards-compatibility.

        Returns:
            django.db.models.query.QuerySet:
            The resulting queryset.
        """
        q = Q()

        if user.is_superuser:
            if visible_only:
                q &= Q(visible=True)
        else:
            q &= Q(public=True)

            if visible_only:
                # We allow accessible() to return hidden repositories if the
                # user is a member, so we must perform this check here.
                q &= Q(visible=True)

            if user.is_authenticated:
                q |= (Q(users__pk=user.pk) |
                      Q(review_groups__users=user.pk))

        if local_site is not LocalSite.ALL:
            q &= Q(local_site=local_site)

        queryset = self.filter(q)

        if distinct:
            queryset = queryset.distinct()

        return queryset

    def accessible_ids(self, *args, **kwargs) -> Sequence[int]:
        """Return IDs of repositories that are accessible by the given user.

        This wraps :py:meth:`accessible` and takes the same arguments
        (with the exception of ``distinct``, which is ignored).

        Callers should not assume order.

        Version Changed:
            3.0.24:
            In prior versions, the order was not specified, but was
            generally numeric order. This should still be true, but
            officially, we no longer guarantee any order of results.

        Args:
            *args (tuple):
                Positional arguments to pass to :py:meth:`accessible`.

            **kwargs (dict):
                Keyword arguments to pass to :py:meth:`accessible`.

        Returns:
            list of int:
            The list of IDs.
        """
        kwargs['distinct'] = False

        return list(sorted(set(
            self.accessible(*args, **kwargs)
            .values_list('pk', flat=True)
        )))

    def get_best_match(
        self,
        repo_identifier: str,
        local_site: Optional[LocalSite] = None,
    ) -> Repository:
        """Return a repository best matching the provided identifier.

        This is used when a consumer provides a repository identifier of
        some form (database ID, repository name, path, or mirror path), and
        needs a single repository back.

        If the identifier appears to be a numeric database ID, then an attempt
        will be made to fetch based on that ID.

        Otherwise, this will perform a lookup, fetching all repositories where
        a name, path, or mirror path match the identifier. If multiple entries
        are found, the following checks are performed in order to locate a
        match:

        1. Is there a single visible repository? If so, return that.
        2. Is there a single repository name matching the identifier? If so,
           return that.
        3. Is there a single repository with a path matching the identifier?
           If so, return that.
        4. Is there a single repository with a mirror path matching the
           identifier? If so, return that.

        Anything else will cause the lookup to fail.

        Note:
           This does not check whether a user has access to the repository.
           Access does not influence any of the checks. The caller is expected
           to check for access permissions using
           :py:meth:`~reviewboard.scmtools.models.Repository.is_accessible_by`
           on the resulting repository.

        Args:
            repo_identifier (str):
                An identifier used to look up a repository.

            local_site (reviewboard.site.models.LocalSite, optional):
                An optional :term:`Local Site` to restrict repositories to.

        Returns:
            reviewboard.scmtools.models.Repository:
            The resulting repository, if one is found.

        Raises:
            reviewboard.scmtools.models.Repository.DoesNotExist:
                A repository could not be found.

            reviewboard.scmtools.models.Repository.MultipleObjectsReturned:
                Too many repositories matching the identifier were found.
        """
        try:
            repo_pk = int(repo_identifier)
        except ValueError:
            repo_pk = None

        if repo_pk is not None:
            # This may raise DoesNotExist.
            return self.get(pk=int(repo_identifier),
                            local_site=local_site)

        # The repository is not a model ID. Try to find one or more
        # repositories with the identifier as a name, path, or mirror path.
        repositories = list(self.filter(
            (Q(path=repo_identifier) |
             Q(mirror_path=repo_identifier) |
             Q(name=repo_identifier)) &
            Q(local_site=local_site)))

        if not repositories:
            # No repository was found matching the identifier provided.
            raise self.model.DoesNotExist

        if len(repositories) == 1:
            # We found an exact match. Return that.
            return repositories[0]

        # Several possible repositories were returned. Let's figure out if
        # one of them is a better candidate than the others.
        #
        # First, check if there's a single match that's visible. We'll choose
        # that one.
        visible_repositories = [
            _repository
            for _repository in repositories
            if _repository.visible
        ]

        if len(visible_repositories) == 1:
            # We found one visible match. Return that.
            return visible_repositories[0]

        # We found either no repositories, or more than 1. Next, see if
        # this matches any explicitly by name.
        named_repositories = [
            _repository
            for _repository in repositories
            if _repository.name == repo_identifier
        ]

        if len(named_repositories) == 1:
            # We found one repository with this name. Return that.
            return named_repositories[0]

        # Last, we'll prioritize paths over mirror paths.
        path_repositories = [
            _repository
            for _repository in repositories
            if _repository.path == repo_identifier
        ]

        if len(path_repositories) == 1:
            # We found one repository with this as the path. Return that.
            return path_repositories[0]

        # We couldn't find a match. It's up to the caller to be more specific.
        raise self.model.MultipleObjectsReturned

    def can_create(
        self,
        user: Union[AnonymousUser, User],
        local_site: Optional[LocalSite] = None,
    ) -> bool:
        """Return whether a user can add a repository.

        Args:
            user (django.contrib.auth.models.User):
                The user to check for repository creation permissions.

            local_site (reviewboard.site.models.LocalSite, optional):
                The optional Local Site that the repository would be added to.
        """
        return user.has_perm('scmtools.add_repository', local_site)

    def encrypt_plain_text_passwords(self) -> None:
        """Encrypts any stored plain-text passwords."""
        qs = self.exclude(
            Q(encrypted_password=None) |
            Q(encrypted_password='') |
            Q(encrypted_password__startswith=
              self.model.ENCRYPTED_PASSWORD_PREFIX))
        qs = qs.only('encrypted_password')

        for repository in qs:
            # This will trigger a migration of the password.
            repository.password



"""View for sending data for updating the review request page."""

import io
import json
import logging
import struct
from datetime import timezone
from typing import Optional

import dateutil.parser
from django.conf import settings
from django.http import (Http404,
                         HttpRequest,
                         HttpResponse,
                         HttpResponseBadRequest)
from django.template.loader import render_to_string
from django.utils.timezone import is_aware, make_aware
from django.views.generic.base import ContextMixin, View
from djblets.util.serializers import DjbletsJSONEncoder
from djblets.views.generic.etag import ETagViewMixin

from reviewboard.reviews.context import make_review_request_context
from reviewboard.reviews.detail import ReviewRequestPageData, entry_registry
from reviewboard.reviews.markdown_utils import is_rich_text_default_for_user
from reviewboard.reviews.views.mixins import ReviewRequestViewMixin


logger = logging.getLogger(__name__)


class ReviewRequestUpdatesView(ReviewRequestViewMixin, ETagViewMixin,
                               ContextMixin, View):
    """Internal view for sending data for updating the review request page.

    This view serializes data representing components of the review request
    page (the issue summary table and entries) that need to periodically
    update without a full page reload. It's used internally by the page to
    request and handle updates.

    The resulting format is a custom, condensed format containing metadata
    information and HTML for each component being updated. It's designed
    to be quick to parse and reduces the amount of data to send across the
    wire (unlike a format like JSON, which would add overhead to the
    serialization/deserialization time and data size when storing HTML).

    Each entry in the payload is in the following format, with all entries
    joined together:

        <metadata length>\\n
        <metadata content>
        <html length>\\n
        <html content>

    The format is subject to change without notice, and should not be
    relied upon by third parties.
    """

    def __init__(
        self,
        **kwargs,
    ) -> None:
        """Initialize the view.

        Args:
            **kwargs (tuple):
                Keyword arguments passed to :py:meth:`as_view`.
        """
        super().__init__(**kwargs)

        self.entry_ids = {}
        self.data = None
        self.since = None

    def pre_dispatch(
        self,
        request: HttpRequest,
        *args,
        **kwargs
    ) -> Optional[HttpResponse]:
        """Look up and validate state before dispatching the request.

        This looks up information based on the request before performing any
        ETag generation or otherwise handling the HTTP request.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            *args (tuple, unused):
                Positional arguments passed to the view.

            **kwargs (dict, unused):
                Keyword arguments passed to the view.

        Returns:
            django.http.HttpResponse:
            The HTTP response containing the updates payload.

        Raises:
            Http404:
                The entry state could not be loaded.
        """
        super().pre_dispatch(request, *args, **kwargs)

        # Find out which entries and IDs (if any) that the caller is most
        # interested in.
        entries_str = request.GET.get('entries')

        if entries_str:
            try:
                for entry_part in entries_str.split(';'):
                    entry_type, entry_ids = entry_part.split(':')
                    self.entry_ids[entry_type] = set(entry_ids.split(','))
            except ValueError as e:
                return HttpResponseBadRequest('Invalid ?entries= value: %s'
                                              % e)

        if self.entry_ids:
            entry_classes = []

            for entry_type in self.entry_ids.keys():
                entry_cls = entry_registry.get_entry(entry_type)

                if entry_cls:
                    entry_classes.append(entry_cls)
        else:
            entry_classes = list(entry_registry)

        if not entry_classes:
            raise Http404

        self.since = request.GET.get('since')

        self.data = ReviewRequestPageData(self.review_request, request,
                                          entry_classes=entry_classes)

    def get_etag_data(
        self,
        request: HttpRequest,
        *args,
        **kwargs,
    ) -> str:
        """Return an ETag for the view.

        This will look up state needed for the request and generate a
        suitable ETag. Some of the information will be stored for later
        computation of the payload.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            *args (tuple, unused):
                Positional arguments passed to the handler.

            **kwargs (dict, unused):
                Keyword arguments passed to the handler.

        Returns:
            str:
            The ETag for the page.
        """
        review_request = self.review_request
        data = self.data
        assert data is not None

        # Build page data only for the entry we care about.
        data.query_data_pre_etag()

        last_activity_time = review_request.get_last_activity_info(
            data.diffsets, data.reviews)['timestamp']

        entry_etags = ':'.join(
            entry_cls.build_etag_data(data)
            for entry_cls in entry_registry
        )

        return ':'.join(str(value) for value in (
            request.user,
            last_activity_time,
            data.latest_review_timestamp,
            review_request.last_review_activity_timestamp,
            entry_etags,
            is_rich_text_default_for_user(request.user),
            settings.AJAX_SERIAL,
        ))

    def get(
        self,
        request: HttpRequest,
        **kwargs,
    ) -> HttpResponse:
        """Handle HTTP GET requests for this view.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            **kwargs (dict):
                Keyword arguments passed to the handler.

        Returns:
            django.http.HttpResponse:
            The HTTP response to send to the client. This will contain the
            custom update payload content.
        """
        request = self.request
        review_request = self.review_request
        data = self.data
        since = self.since

        assert data is not None

        # Finish any querying needed by entries on this page.
        data.query_data_post_etag()

        # Gather all the entries into a single list.
        #
        # Note that the order in which we build the resulting list of entries
        # doesn't matter at this stage, but it does need to be consistent.
        # The current order (main, initial) is based on Python 2.7 sort order,
        # which our tests are based on. This could be changed in the future.
        all_entries = data.get_entries()
        entries = all_entries['main'] + all_entries['initial']

        if self.entry_ids:
            # If specific entry IDs have been requested, limit the results
            # to those.
            entries = (
                entry
                for entry in entries
                if (entry.entry_type_id in self.entry_ids and
                    entry.entry_id in self.entry_ids[entry.entry_type_id])
            )

        # See if the caller only wants to fetch entries updated since a given
        # timestamp.
        if since:
            since = dateutil.parser.parse(since)

            if not is_aware(since):
                since = make_aware(since, timezone.utc)

            entries = (
                entry
                for entry in entries
                if (entry.updated_timestamp is not None and
                    entry.updated_timestamp.replace(microsecond=0) > since)
            )

        # We can now begin to serialize the payload for all the updates.
        payload = io.BytesIO()
        base_entry_context = None
        needs_issue_summary_table = False

        for entry in entries:
            metadata = {
                'type': 'entry',
                'entryType': entry.entry_type_id,
                'entryID': entry.entry_id,
                'etag': entry.build_etag_data(data, entry=entry),
                'addedTimestamp': entry.added_timestamp,
                'updatedTimestamp': entry.updated_timestamp,
                'modelData': entry.get_js_model_data(),
                'viewOptions': entry.get_js_view_data(),
            }

            if base_entry_context is None:
                # Now that we know the context is needed for entries,
                # we can construct and populate it.
                base_entry_context = super().get_context_data(**kwargs)
                base_entry_context.update(
                    make_review_request_context(
                        request=request,
                        review_request=review_request))

            try:
                html = render_to_string(
                    template_name=entry.template_name,
                    context=dict({
                        'show_entry_statuses_area': (
                            entry.entry_pos == entry.ENTRY_POS_MAIN),
                        'entry': entry,
                    }, **base_entry_context),
                    request=request)

                self._write_update(payload, metadata, html)
            except Exception as e:
                logger.error('Error rendering review request page entry '
                             '%r: %s',
                             entry, e,
                             extra={'request': request})

            if entry.needs_reviews:
                needs_issue_summary_table = True

        # If any of the entries required any information on reviews, then
        # the state of the issue summary table may have changed. We'll need
        # to send this along as well.
        if (needs_issue_summary_table and
            (since is None or
             data.latest_issue_timestamp.replace(microsecond=0) > since)):
            metadata = {
                'type': 'issue-summary-table',
                'updatedTimestamp': data.latest_issue_timestamp,
            }

            html = render_to_string(
                template_name='reviews/review_issue_summary_table.html',
                context={
                    'issue_counts': data.issue_counts,
                    'issues': data.issues,
                },
                request=request)

            self._write_update(payload, metadata, html)

        # The payload's complete. Close it out and send to the client.
        result = payload.getvalue()
        payload.close()

        return HttpResponse(result, content_type='text/plain; charset=utf-8')

    def _write_update(
        self,
        payload: io.BytesIO,
        metadata: dict,
        html: str,
    ) -> None:
        """Write an update to the payload.

        This will format the metadata and HTML for the update and write it.

        Args:
            payload (io.BytesIO):
                The payload to write to.

            metadata (dict):
                The JSON-serializable metadata to write.

            html (str):
                The HTML to write.
        """
        metadata_bytes = (
            json.dumps(metadata,
                       cls=DjbletsJSONEncoder,
                       sort_keys=True)
            .encode('utf-8')
        )
        html_bytes = html.strip().encode('utf-8')

        payload.write(struct.pack(b'<L', len(metadata_bytes)))
        payload.write(metadata_bytes)
        payload.write(struct.pack(b'<L', len(html_bytes)))
        payload.write(html_bytes)



"""Unit tests for reviewboard.licensing.license.LicenseInfo.

Version Added:
    8.0
"""

from __future__ import annotations

from datetime import datetime, timedelta

import kgb
from django.utils import timezone

from reviewboard.licensing.license import LicenseInfo, LicenseStatus
from reviewboard.testing import TestCase


class LicenseInfoTests(kgb.SpyAgency, TestCase):
    """Unit tests for LicenseInfo.

    Version Added:
        8.0
    """

    def setUp(self) -> None:
        """Set up a unit test.

        This will set up a spy to ensure :py:func:`django.utils.timezone.now`
        always returns a static date/time.
        """
        super().setUp()

        self.spy_on(
            timezone.now,
            op=kgb.SpyOpReturn(timezone.make_aware(datetime(2025, 4, 19))))

    def test_get_expires_soon_with_unlicensed(self) -> None:
        """Testing LicenseInfo.get_expires_soon with unlicensed"""
        license_info = LicenseInfo(
            expires=timezone.now() + timedelta(days=1),
            license_id='test-license',
            licensed_to='Test User',
            product_name='Test Product',
            status=LicenseStatus.UNLICENSED)

        self.assertFalse(license_info.get_expires_soon())

    def test_get_expires_soon_with_no_expires(self) -> None:
        """Testing LicenseInfo.get_expires_soon without an expiration"""
        license_info = LicenseInfo(
            expires=None,
            license_id='test-license',
            licensed_to='Test User',
            product_name='Test Product',
            status=LicenseStatus.LICENSED)

        self.assertFalse(license_info.get_expires_soon())

    def test_get_expires_soon_with_trial_10_days(self) -> None:
        """Testing LicenseInfo.get_expires_soon with trial and expires in 10
        days
        """
        license_info = LicenseInfo(
            expires=timezone.now() + timedelta(days=10),
            is_trial=True,
            license_id='test-license',
            licensed_to='Test User',
            product_name='Test Product',
            status=LicenseStatus.LICENSED)

        self.assertTrue(license_info.get_expires_soon())

    def test_get_expires_soon_with_trial_11_days(self) -> None:
        """Testing LicenseInfo.get_expires_soon with trial and expires in 11
        days
        """
        license_info = LicenseInfo(
            expires=timezone.now() + timedelta(days=11),
            is_trial=True,
            license_id='test-license',
            licensed_to='Test User',
            product_name='Test Product',
            status=LicenseStatus.LICENSED)

        self.assertFalse(license_info.get_expires_soon())

    def test_get_expires_soon_with_purchased_30_days(self) -> None:
        """Testing LicenseInfo.get_expires_soon with purchased license and
        30 days
        """
        license_info = LicenseInfo(
            expires=timezone.now() + timedelta(days=30),
            license_id='test-license',
            licensed_to='Test User',
            product_name='Test Product',
            status=LicenseStatus.LICENSED)

        self.assertTrue(license_info.get_expires_soon())

    def test_get_expires_soon_with_purchased_31_days(self) -> None:
        """Testing LicenseInfo.get_expires_soon with purchased license and
        31 days
        """
        license_info = LicenseInfo(
            expires=timezone.now() + timedelta(days=31),
            license_id='test-license',
            licensed_to='Test User',
            product_name='Test Product',
            status=LicenseStatus.LICENSED)

        self.assertFalse(license_info.get_expires_soon())

    def test_get_summary_with_custom_summary(self) -> None:
        """Testing LicenseInfo.get_summary with custom summary"""
        license_info = LicenseInfo(
            product_name='Test Product',
            license_id='test-license',
            licensed_to='Test User',
            summary='This is my custom summary')

        self.assertEqual(license_info.get_summary(),
                         'This is my custom summary')

    def test_get_summary_with_unlicensed(self) -> None:
        """Testing LicenseInfo.get_summary with unlicensed state"""
        license_info = LicenseInfo(
            product_name='Test Product',
            license_id='test-license',
            licensed_to='Test User',
            status=LicenseStatus.UNLICENSED)

        self.assertEqual(license_info.get_summary(),
                         'Test Product is not licensed!')

    def test_get_summary_with_licensed(self) -> None:
        """Testing LicenseInfo.get_summary with purchased license"""
        license_info = LicenseInfo(
            product_name='Test Product',
            license_id='test-license',
            licensed_to='Test User',
            status=LicenseStatus.LICENSED)

        self.assertEqual(license_info.get_summary(),
                         'License for Test Product is active')

    def test_get_summary_with_licensed_plan(self) -> None:
        """Testing LicenseInfo.get_summary with purchased license with plan"""
        license_info = LicenseInfo(
            product_name='Test Product',
            license_id='test-license',
            licensed_to='Test User',
            plan_name='Super Plan',
            status=LicenseStatus.LICENSED)

        self.assertEqual(license_info.get_summary(),
                         'License for Test Product Super Plan is active')

    def test_get_summary_with_licensed_expires_soon(self) -> None:
        """Testing LicenseInfo.get_summary with purchased license expires
        soon
        """
        license_info = LicenseInfo(
            expires=timezone.now() + timedelta(days=10),
            product_name='Test Product',
            license_id='test-license',
            licensed_to='Test User',
            status=LicenseStatus.LICENSED)

        self.assertEqual(license_info.get_summary(),
                         'License for Test Product expires in 10 days')

    def test_get_summary_with_licensed_plan_expires_soon(self) -> None:
        """Testing LicenseInfo.get_summary with purchased license with plan
        expires soon
        """
        license_info = LicenseInfo(
            expires=timezone.now() + timedelta(days=10),
            product_name='Test Product',
            license_id='test-license',
            licensed_to='Test User',
            plan_name='Super Plan',
            status=LicenseStatus.LICENSED)

        self.assertEqual(license_info.get_summary(),
                         'License for Test Product Super Plan expires in '
                         '10 days')

    def test_get_summary_with_licensed_auto_renew(self) -> None:
        """Testing LicenseInfo.get_summary with purchased license set to
        auto-renew within the expires-soon window
        """
        license_info = LicenseInfo(
            auto_renew=True,
            expires=timezone.now() + timedelta(days=10),
            product_name='Test Product',
            license_id='test-license',
            licensed_to='Test User',
            status=LicenseStatus.LICENSED)

        self.assertEqual(license_info.get_summary(),
                         'License for Test Product renews in 10 days')

    def test_get_summary_with_licensed_plan_auto_renew(self) -> None:
        """Testing LicenseInfo.get_summary with purchased license with plan
        set to auto-renew within the expires-soon window
        """
        license_info = LicenseInfo(
            auto_renew=True,
            expires=timezone.now() + timedelta(days=10),
            product_name='Test Product',
            license_id='test-license',
            licensed_to='Test User',
            plan_name='Super Plan',
            status=LicenseStatus.LICENSED)

        self.assertEqual(license_info.get_summary(),
                         'License for Test Product Super Plan renews in '
                         '10 days')

    def test_get_summary_with_trial(self) -> None:
        """Testing LicenseInfo.get_summary with trial license"""
        license_info = LicenseInfo(
            expires=timezone.now() + timedelta(days=10),
            product_name='Test Product',
            license_id='test-license',
            licensed_to='Test User',
            is_trial=True,
            status=LicenseStatus.LICENSED)

        self.assertEqual(license_info.get_summary(),
                         'Trial license for Test Product ends in 10 days')

    def test_get_summary_with_trial_1_day_remaining(self) -> None:
        """Testing LicenseInfo.get_summary with trial license with 1 day
        remaining
        """
        license_info = LicenseInfo(
            expires=timezone.now() + timedelta(days=1),
            product_name='Test Product',
            license_id='test-license',
            licensed_to='Test User',
            is_trial=True,
            status=LicenseStatus.LICENSED)

        self.assertEqual(license_info.get_summary(),
                         'Trial license for Test Product ends in 1 day')

    def test_get_summary_with_trial_plan(self) -> None:
        """Testing LicenseInfo.get_summary with trial license with plan"""
        license_info = LicenseInfo(
            expires=timezone.now() + timedelta(days=10),
            product_name='Test Product',
            license_id='test-license',
            licensed_to='Test User',
            plan_name='Super Plan',
            is_trial=True,
            status=LicenseStatus.LICENSED)

        self.assertEqual(license_info.get_summary(),
                         'Trial license for Test Product Super Plan ends '
                         'in 10 days')

    def test_get_summary_with_trial_plan_1_day_remaining(self) -> None:
        """Testing LicenseInfo.get_summary with trial license with plan with
        1 day remaining
        """
        license_info = LicenseInfo(
            expires=timezone.now() + timedelta(days=1),
            product_name='Test Product',
            license_id='test-license',
            licensed_to='Test User',
            plan_name='Super Plan',
            is_trial=True,
            status=LicenseStatus.LICENSED)

        self.assertEqual(license_info.get_summary(),
                         'Trial license for Test Product Super Plan ends '
                         'in 1 day')

    def test_get_summary_with_expired_license(self) -> None:
        """Testing LicenseInfo.get_summary with expired license"""
        license_info = LicenseInfo(
            expires=timezone.now() - timedelta(days=10),
            grace_period_days_remaining=5,
            product_name='Test Product',
            license_id='test-license',
            licensed_to='Test User',
            status=LicenseStatus.EXPIRED_GRACE_PERIOD)

        self.assertEqual(license_info.get_summary(),
                         'License for Test Product expired and will stop '
                         'working in 5 days')

    def test_get_summary_with_expired_license_1_day_remaining(self) -> None:
        """Testing LicenseInfo.get_summary with expired license and 1 day
        remaining
        """
        license_info = LicenseInfo(
            expires=timezone.now() - timedelta(days=10),
            grace_period_days_remaining=1,
            product_name='Test Product',
            license_id='test-license',
            licensed_to='Test User',
            status=LicenseStatus.EXPIRED_GRACE_PERIOD)

        self.assertEqual(license_info.get_summary(),
                         'License for Test Product expired and will stop '
                         'working in 1 day')

    def test_get_summary_with_expired_license_plan(self) -> None:
        """Testing LicenseInfo.get_summary with expired license with plan"""
        license_info = LicenseInfo(
            expires=timezone.now() - timedelta(days=10),
            grace_period_days_remaining=5,
            product_name='Test Product',
            license_id='test-license',
            licensed_to='Test User',
            plan_name='Super Plan',
            status=LicenseStatus.EXPIRED_GRACE_PERIOD)

        self.assertEqual(license_info.get_summary(),
                         'License for Test Product Super Plan expired '
                         'and will stop working in 5 days')

    def test_get_summary_with_expired_license_plan_1_day_remaining(
        self,
    ) -> None:
        """Testing LicenseInfo.get_summary with expired license with plan"""
        license_info = LicenseInfo(
            expires=timezone.now() - timedelta(days=10),
            grace_period_days_remaining=1,
            product_name='Test Product',
            license_id='test-license',
            licensed_to='Test User',
            plan_name='Super Plan',
            status=LicenseStatus.EXPIRED_GRACE_PERIOD)

        self.assertEqual(license_info.get_summary(),
                         'License for Test Product Super Plan expired '
                         'and will stop working in 1 day')

    def test_get_summary_with_expired_trial(self) -> None:
        """Testing LicenseInfo.get_summary with expired trial license"""
        license_info = LicenseInfo(
            expires=timezone.now() - timedelta(days=10),
            grace_period_days_remaining=5,
            product_name='Test Product',
            license_id='test-license',
            licensed_to='Test User',
            is_trial=True,
            status=LicenseStatus.EXPIRED_GRACE_PERIOD)

        self.assertEqual(license_info.get_summary(),
                         'Trial license for Test Product expired and will '
                         'stop working in 5 days')

    def test_get_summary_with_expired_trial_1_day_remaining(self) -> None:
        """Testing LicenseInfo.get_summary with expired trial license with
        1 day remaining
        """
        license_info = LicenseInfo(
            expires=timezone.now() - timedelta(days=10),
            grace_period_days_remaining=1,
            product_name='Test Product',
            license_id='test-license',
            licensed_to='Test User',
            is_trial=True,
            status=LicenseStatus.EXPIRED_GRACE_PERIOD)

        self.assertEqual(license_info.get_summary(),
                         'Trial license for Test Product expired and will '
                         'stop working in 1 day')

    def test_get_summary_with_expired_trial_plan(self) -> None:
        """Testing LicenseInfo.get_summary with expired trial license with
        plan
        """
        license_info = LicenseInfo(
            expires=timezone.now() - timedelta(days=10),
            grace_period_days_remaining=5,
            product_name='Test Product',
            license_id='test-license',
            licensed_to='Test User',
            plan_name='Super Plan',
            is_trial=True,
            status=LicenseStatus.EXPIRED_GRACE_PERIOD)

        self.assertEqual(license_info.get_summary(),
                         'Trial license for Test Product Super Plan '
                         'expired and will stop working in 5 days')

    def test_get_summary_with_expired_trial_plan_1_day_remaining(
        self,
    ) -> None:
        """Testing LicenseInfo.get_summary with expired trial license with
        plan and 1 day remaining
        """
        license_info = LicenseInfo(
            expires=timezone.now() - timedelta(days=10),
            grace_period_days_remaining=1,
            product_name='Test Product',
            license_id='test-license',
            licensed_to='Test User',
            plan_name='Super Plan',
            is_trial=True,
            status=LicenseStatus.EXPIRED_GRACE_PERIOD)

        self.assertEqual(license_info.get_summary(),
                         'Trial license for Test Product Super Plan '
                         'expired and will stop working in 1 day')

    def test_get_summary_with_hard_expired_license(self) -> None:
        """Testing LicenseInfo.get_summary with hard-expired license"""
        license_info = LicenseInfo(
            expires=timezone.now() - timedelta(days=10),
            grace_period_days_remaining=5,
            product_name='Test Product',
            license_id='test-license',
            licensed_to='Test User',
            status=LicenseStatus.HARD_EXPIRED)

        self.assertEqual(license_info.get_summary(),
                         'License for Test Product expired and needs to be '
                         'renewed')

    def test_get_summary_with_hard_expired_license_plan(self) -> None:
        """Testing LicenseInfo.get_summary with hard-expired license with plan
        """
        license_info = LicenseInfo(
            expires=timezone.now() - timedelta(days=10),
            grace_period_days_remaining=5,
            product_name='Test Product',
            license_id='test-license',
            licensed_to='Test User',
            plan_name='Super Plan',
            status=LicenseStatus.HARD_EXPIRED)

        self.assertEqual(license_info.get_summary(),
                         'License for Test Product Super Plan expired '
                         'and needs to be renewed')

    def test_get_summary_with_hard_expired_trial(self) -> None:
        """Testing LicenseInfo.get_summary with hard-expired trial license"""
        license_info = LicenseInfo(
            expires=timezone.now() - timedelta(days=10),
            grace_period_days_remaining=5,
            product_name='Test Product',
            license_id='test-license',
            licensed_to='Test User',
            is_trial=True,
            status=LicenseStatus.HARD_EXPIRED)

        self.assertEqual(license_info.get_summary(),
                         'Trial license for Test Product expired')

    def test_get_summary_with_hard_expired_trial_plan(self) -> None:
        """Testing LicenseInfo.get_summary with hard-expired trial license
        with plan
        """
        license_info = LicenseInfo(
            expires=timezone.now() - timedelta(days=10),
            grace_period_days_remaining=5,
            product_name='Test Product',
            license_id='test-license',
            licensed_to='Test User',
            plan_name='Super Plan',
            is_trial=True,
            status=LicenseStatus.HARD_EXPIRED)

        self.assertEqual(license_info.get_summary(),
                         'Trial license for Test Product Super Plan '
                         'expired')



"""API Resource for file attachments associated with files in diffs."""

from __future__ import annotations

from typing import Any, Dict, Optional, TYPE_CHECKING, Union

from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Q
from djblets.util.decorators import augment_method_from
from djblets.webapi.decorators import (webapi_login_required,
                                       webapi_request_fields,
                                       webapi_response_errors)
from djblets.webapi.errors import (DOES_NOT_EXIST,
                                   DUPLICATE_ITEM,
                                   INVALID_FORM_DATA,
                                   NOT_LOGGED_IN,
                                   PERMISSION_DENIED,
                                   WebAPIError)
from djblets.webapi.fields import (BooleanFieldType,
                                   IntFieldType,
                                   ResourceFieldType,
                                   StringFieldType)

from reviewboard.attachments.errors import FileTooBigError
from reviewboard.attachments.forms import UploadFileForm
from reviewboard.attachments.models import FileAttachment
from reviewboard.diffviewer.models import FileDiff
from reviewboard.webapi.base import ImportExtraDataError
from reviewboard.webapi.decorators import webapi_check_local_site
from reviewboard.webapi.errors import DIFF_TOO_BIG
from reviewboard.webapi.resources import resources
from reviewboard.webapi.resources.base_review_request_file_attachment import \
    BaseReviewRequestFileAttachmentResource

if TYPE_CHECKING:
    from django.http import HttpRequest


class DiffFileAttachmentResource(BaseReviewRequestFileAttachmentResource):
    """Provides information on file attachments associated with files in diffs.

    The list of file attachments are tied to files either committed to the
    repository or proposed in diffs to a review request on the repository.
    All are associated with a file in a diff.

    Files that are newly introduced in a diff and do not have a revision as
    of that diff will have the ``added_in_filediff`` link set, and
    ``repository_revision`` will be null.

    Version Changed:
        7.0:
        Added support for HTTP POST to the list resource.
    """

    added_in = '2.0'

    name = 'diff_file_attachment'
    model_parent_key = 'repository'
    allowed_methods = ('GET', 'POST')

    mimetype_list_resource_name = 'diff-file-attachments'
    mimetype_item_resource_name = 'diff-file-attachment'

    fields = dict({
        'repository_file_path': {
            'type': StringFieldType,
            'description': 'The file path inside the repository that this '
                           'file attachment represents.',
        },
        'repository_revision': {
            'type': StringFieldType,
            'description': 'The revision that introduced this version of the '
                           'file, if committed in the repository.',
        },
        'added_in_filediff': {
            'type': ResourceFieldType,
            'resource': 'reviewboard.webapi.resources.filediff.'
                        'FileDiffResource',
            'description': 'The file diff that introduced this file. If set, '
                           'this file is just part of a proposed change, and '
                           'not necessarily committed in the repository.',
        },
    }, **BaseReviewRequestFileAttachmentResource.fields)

    def serialize_repository_file_path_field(self, attachment, **kwargs):
        if attachment.added_in_filediff_id:
            return attachment.added_in_filediff.dest_file
        else:
            return attachment.repo_path

    def serialize_repository_revision_field(self, attachment, **kwargs):
        return attachment.repo_revision or None

    def has_access_permissions(self, request, obj, *args, **kwargs):
        repository = self.get_parent_object(obj)

        return repository.is_accessible_by(request.user)

    def get_queryset(self, request, is_list=False, *args, **kwargs):
        repository = resources.repository.get_object(request, *args, **kwargs)
        queryset = self.model.objects.filter_for_repository(repository)

        if is_list:
            q = Q()

            if 'repository-file-path' in kwargs:
                path = kwargs['repository-file-path']

                q = q & (Q(repo_path=path) |
                         Q(added_in_filediff__source_file=path))

            if 'repository-revision' in kwargs:
                q = q & Q(repo_revision=kwargs['repository-revision'])

            if 'mimetype' in kwargs:
                q = q & Q(mimetype=kwargs['mimetype'])

            queryset = queryset.filter(q)

        return queryset

    def get_parent_object(self, obj):
        if obj.repository_id is None:
            assert obj.added_in_filediff_id is not None
            return obj.added_in_filediff.get_repository()
        else:
            return obj.repository

    @webapi_request_fields(
        optional=dict({
            'repository-file-path': {
                'type': StringFieldType,
                'description': (
                    'Filter file attachments with the given path in the '
                    'repository.'
                ),
            },
            'repository-revision': {
                'type': StringFieldType,
                'description': (
                    'Filter file attachments for files with the given '
                    'revision.'
                ),
            },
            'mimetype': {
                'type': StringFieldType,
                'description': (
                    'Filter file attachments with the given mimetype.'
                ),
            },
        }, **BaseReviewRequestFileAttachmentResource.get_list.optional_fields),
        required=BaseReviewRequestFileAttachmentResource.get_list
                                                        .required_fields
    )
    @augment_method_from(BaseReviewRequestFileAttachmentResource)
    def get_list(self, request, *args, **kwargs):
        """Returns the list of file attachments associated with diffs.

        Each item in this list is a file attachment associated with a file
        or a proposed change against the parent repository. A file attachment
        may be referenced by one or more diffs.
        """
        pass

    @webapi_check_local_site
    @webapi_login_required
    @webapi_response_errors(DOES_NOT_EXIST, PERMISSION_DENIED,
                            INVALID_FORM_DATA, NOT_LOGGED_IN)
    @webapi_request_fields(
        required={
            'filediff': {
                'type': IntFieldType,
                'description': 'The ID of the FileDiff to create the '
                               'attachment for.',
            },
        },
        optional={
            'source_file': {
                'type': BooleanFieldType,
                'description': 'Whether to create the attachment for the '
                               'source version of the file. If false or not '
                               'specified, the attachment will be created '
                               'for the modified version.',
                'added_in': '7.0',
            },
        },
        allow_unknown=True,
    )
    def create(
        self,
        request: HttpRequest,
        filediff: Optional[int] = None,
        source_file: Optional[bool] = None,
        extra_fields: Dict[str, Any] = {},
        *args,
        **kwargs,
    ) -> Union[tuple, WebAPIError]:
        """Creates a new diff file attachment.

        This will create a new file attachment and connect it to the provided
        FileDiff, for use with binary files in diffs.
        """
        try:
            repository = \
                resources.repository.get_object(request, *args, **kwargs)
        except ObjectDoesNotExist:
            return DOES_NOT_EXIST

        if not repository.is_accessible_by(request.user):
            return self.get_no_access_error(request)

        try:
            filediff_obj = FileDiff.objects.get(pk=filediff)
        except ObjectDoesNotExist as e:
            return INVALID_FORM_DATA, {
                'fields': {
                    'filediff': [str(e)],
                },
            }

        review_request = filediff_obj.get_review_request()

        if not review_request.is_mutable_by(request.user):
            return self.get_no_access_error(request)

        modified = source_file is not True

        existing_attachment = FileAttachment.objects.get_for_filediff(
            filediff_obj, modified)

        if existing_attachment is not None:
            return DUPLICATE_ITEM

        form_data = request.POST.copy()
        form = UploadFileForm(review_request, form_data, request.FILES)

        if not form.is_valid():
            return INVALID_FORM_DATA, {
                'fields': self._get_form_errors(form),
            }

        try:
            file = form.create(filediff=filediff_obj,
                               from_modified=modified)
        except FileTooBigError as e:
            return DIFF_TOO_BIG, {
                'max_size': e.max_attachment_size,
                'reason': str(e),
            }
        except ValueError as e:
            return INVALID_FORM_DATA, {
                'fields': {
                    'path': [str(e)],
                },
            }

        if extra_fields:
            try:
                self.import_extra_data(file, file.extra_data, extra_fields)
            except ImportExtraDataError as e:
                return e.error_payload

            file.save(update_fields=('extra_data',))

        return 201, {
            self.item_result_key: self.serialize_object(
                file, request=request, *args, **kwargs),
        }


diff_file_attachment_resource = DiffFileAttachmentResource()



from django.conf import settings
from djblets.siteconfig.models import SiteConfiguration

from reviewboard import (get_manual_url, get_package_version,
                         get_version_string, is_release, VERSION)
from reviewboard.admin.read_only import is_site_read_only_for


def read_only(request):
    """Return a dictionary with the read-only state.

    Args:
        request (django.http.HttpRequest):
            The current HTTP request.

    Returns:
        dict:
        State to add to the context.
    """
    siteconfig = SiteConfiguration.objects.get_current()

    return {
        'is_read_only': is_site_read_only_for(request.user),
        'read_only_message': siteconfig.get('read_only_message', ''),
    }


def version(request):
    """Return a dictionary with version information.

    Args:
        request (django.http.HttpRequest):
            The current HTTP request.

    Returns:
        dict:
        State to add to the context.
    """
    return {
        'version': get_version_string(),
        'package_version': get_package_version(),
        'is_release': is_release(),
        'version_raw': VERSION,
        'PRODUCT_NAME': settings.PRODUCT_NAME,
        'RB_MANUAL_URL': get_manual_url(),
    }



"""Administration form for search settings."""

from django import forms
from django.core.exceptions import ValidationError
from django.utils.translation import gettext, gettext_lazy as _
from djblets.siteconfig.forms import SiteSettingsForm

from reviewboard.admin.siteconfig import load_site_config
from reviewboard.search import search_backend_registry


class SearchSettingsForm(SiteSettingsForm):
    """Form for search settings.

    This form manages the main search settings (enabled, how many results, and
    what backend to use), as well as displaying per-search backend forms so
    that they may be configured.

    For example, Elasticsearch requires a URL and index name, while Whoosh
    requires a file path to store its index. These fields (and fields for any
    other added search backend) will only be shown to the user when the
    appropriate search backend is selected.
    """

    search_enable = forms.BooleanField(
        label=_('Enable search'),
        help_text=_('If enabled, provides a search field for quickly '
                    'searching through review requests, diffs, and users.'),
        required=False)

    search_results_per_page = forms.IntegerField(
        label=_('Search results per page'),
        min_value=1,
        required=False)

    search_backend_id = forms.ChoiceField(
        label=_('Search backend'),
        required=False,
        widget=forms.Select(attrs={
            'data-subform-group': 'search-backend',
        }))

    search_on_the_fly_indexing = forms.BooleanField(
        label=_('On-the-fly indexing'),
        required=False,
        help_text=('If enabled, the search index will be updated dynamically '
                   'when review requests or users change.<br>'
                   '<strong>Note:</strong> This is not recommended for use '
                   'with the Whoosh engine for large or multi-server '
                   'installs.'))

    def __init__(self, siteconfig, data=None, *args, **kwargs):
        """Initialize the search engine settings form.

        This will also initialize the settings forms for each search engine
        backend.

        Args:
            site_config (djblets.siteconfig.models.SiteConfiguration):
                The site configuration handling the server's settings.

            data (dict, optional):
                The form data.

            *args (tuple):
                Additional positional arguments.

            **kwargs (dict):
                Additional keyword arguments.
        """
        super(SearchSettingsForm, self).__init__(siteconfig, data, *args,
                                                 **kwargs)

        request = kwargs.get('request')
        cur_search_backend_id = (self['search_backend_id'].data or
                                 self.fields['search_backend_id'].initial)

        choices = []
        search_backend_forms = {}

        for backend in search_backend_registry:
            search_backend_id = backend.search_backend_id

            if cur_search_backend_id == search_backend_id:
                subform = backend.get_config_form(data,
                                                  files=kwargs.get('files'),
                                                  request=request)
            else:
                subform = backend.get_config_form(request=request)

            search_backend_forms[search_backend_id] = subform
            choices.append((search_backend_id, backend.name))

        self.search_backend_forms = search_backend_forms
        self.fields['search_backend_id'].choices = choices

    def is_valid(self):
        """Return whether the form is valid.

        This will check the validity of the fields on this form and on
        the selected search backend's settings form.

        Returns:
            bool:
            ``True`` if the main settings form and search backend's settings
            form is valid. ``False`` if either form is invalid.
        """
        if not super(SearchSettingsForm, self).is_valid():
            return False

        backend_id = self.cleaned_data['search_backend_id']
        backend_form = self.search_backend_forms[backend_id]

        return backend_form.is_valid()

    def clean_search_backend_id(self):
        """Clean the ``search_backend_id`` field.

        This will ensure the chosen search backend is valid (i.e., it is
        available in the registry) and that its dependencies have been
        installed.

        Returns:
            unicode:
            The search backend ID.

        Raises:
            django.core.exceptions.ValidationError:
                Raised if the search engine ID chosen cannot be used.
        """
        search_backend_id = self.cleaned_data['search_backend_id']
        search_backend = search_backend_registry.get_search_backend(
            search_backend_id)

        if not search_backend:
            raise ValidationError(
                gettext('The search engine "%s" could not be found. '
                        'If this is provided by an extension, you will have '
                        'to make sure that extension is enabled.')
                % search_backend_id
            )

        return search_backend_id

    def clean(self):
        """Clean the form and the sub-form for the selected search backend.

        Returns:
            dict:
            The cleaned data.
        """
        cleaned_data = self.cleaned_data

        if cleaned_data['search_enable']:
            search_backend_id = cleaned_data.get('search_backend_id')

            # The search_backend_id field is only available if the backend
            # passed validation.
            if search_backend_id:
                backend_form = self.search_backend_forms[search_backend_id]

                # Validate the configuration form.
                if backend_form.is_valid():
                    # Validate the search backend, ensuring the configuration
                    # is correct.
                    search_backend = \
                        search_backend_registry.get_search_backend(
                            search_backend_id)
                    configuration = \
                        search_backend.get_configuration_from_form_data(
                            backend_form.cleaned_data)

                    try:
                        search_backend.validate(configuration=configuration)
                    except ValidationError as e:
                        self.add_error('search_backend_id', e.error_list)
                else:
                    # The configuration form had issues.
                    self._errors.update(backend_form.errors)

        return cleaned_data

    def save(self):
        """Save the form and sub-form for the selected search backend.

        This forces a site configuration reload.
        """
        search_backend_id = self.cleaned_data['search_backend_id']

        if self.cleaned_data['search_enable']:
            # We only need to update the backend settings when search is
            # enabled.
            backend_form = self.search_backend_forms[search_backend_id]
            backend = search_backend_registry.get_search_backend(
                search_backend_id)
            backend.configuration = backend.get_configuration_from_form_data(
                backend_form.cleaned_data)

        super(SearchSettingsForm, self).save()

        # Reload any import changes to the Django settings.
        load_site_config()

    class Meta:
        title = _('Search Settings')
        subforms = (
            {
                'subforms_attr': 'search_backend_forms',
                'controller_field': 'search_backend_id',
            },
        )



"""Unit tests for reviewboard.admin.forms.AuthenticationSettingsForm.

Version Added:
    5.0.5
"""

from djblets.siteconfig.models import SiteConfiguration

from reviewboard.admin.forms.auth_settings import \
    AuthenticationSettingsForm
from reviewboard.testing.testcase import TestCase


class AuthenticationSettingsFormTests(TestCase):
    """Unit tests for reviewboard.admin.forms.AuthenticationSettingsForm.

    Version Added:
        5.0.5
    """

    def test_load_with_none(self) -> None:
        """Test AuthenticationSettingsForm.load with no API tokens
        expiration set in siteconfig
        """
        siteconfig_settings = {
            'client_token_expiration': None,
        }

        with self.siteconfig_settings(siteconfig_settings):
            siteconfig = SiteConfiguration.objects.get_current()
            form = AuthenticationSettingsForm(siteconfig)
            form.load()

            self.assertIsNone(form['automatic_api_token_expiration'].value())

    def test_load_with_date(self) -> None:
        """Test AuthenticationSettingsForm.load with API tokens
        expiration set in siteconfig
        """
        siteconfig_settings = {
            'client_token_expiration': 10,
        }

        with self.siteconfig_settings(siteconfig_settings):
            siteconfig = SiteConfiguration.objects.get_current()
            form = AuthenticationSettingsForm(siteconfig)
            form.load()

            self.assertEqual(form['automatic_api_token_expiration'].value(),
                             10)

    def test_save_with_none(self) -> None:
        """Test AuthenticationSettingsForm.save with API tokens
        set to never expire
        """
        siteconfig_settings = {
            'client_token_expiration': 10,
        }

        with self.siteconfig_settings(siteconfig_settings):
            siteconfig = SiteConfiguration.objects.get_current()
            form = AuthenticationSettingsForm(siteconfig, data={
                'auth_backend': 'builtin',
                'automatic_api_token_expiration_0': '20',
                'automatic_api_token_expiration_1': '',
            })

            self.assertTrue(form.is_valid())

            form.save()

            self.assertIsNone(
                siteconfig.get('client_token_expiration'))

    def test_save_with_date(self) -> None:
        """Test AuthenticationSettingsForm.save with API tokens
        expiration set
        """
        siteconfig_settings = {
            'client_token_expiration': None,
        }

        with self.siteconfig_settings(siteconfig_settings):
            siteconfig = SiteConfiguration.objects.get_current()
            form = AuthenticationSettingsForm(siteconfig, data={
                'auth_backend': 'builtin',
                'automatic_api_token_expiration_0': '20',
                'automatic_api_token_expiration_1': '1',
            })

            self.assertTrue(form.is_valid())

            form.save()

            self.assertEqual(
                siteconfig.get('client_token_expiration'),
                20)



"""Errors for actions app.

Version Added:
    6.0
"""

from __future__ import annotations

from housekeeping import ClassDeprecatedMixin

from reviewboard.deprecation import RemovedInReviewBoard10_0Warning


class ActionError(Exception):
    """Base class for action-related errors.

    Version:
        8.0
    """


class MissingActionRendererError(ActionError):
    """An error indicating an action renderer could not be found.

    Version:
        8.0
    """


class DepthLimitExceededError(ClassDeprecatedMixin,
                              ValueError,
                              warning_cls=RemovedInReviewBoard10_0Warning):
    """An error that occurs when the maximum depth limit is exceeded.

    Actions cannot be arbitrarily nested. For example, if the depth limit is 2,
    then this error would be triggered if an extension tried to add a menu
    action as follows:

    .. code-block:: python

       ActionHook(self, actions=[
           DepthZeroMenuAction(),
           DepthOneFirstItemAction(),
           DepthOneMenuAction(),
           DepthTwoMenuAction(),  # This depth is acceptable.
           DepthThreeTooDeepAction(),  # This action is too deep.
       ])

    Deprecated:
        8.0:
        This class is no longer used. It will be removed in Review Board 10.

    Version Added:
        6.0
    """

    def __init__(
        self,
        action_id: str,
        *,
        depth_limit: int,
    ) -> None:
        """Initialize the error.

        Args:
            action_id (str):
                The ID of the action which was too deep.

            depth_limit (int):
                The maximum depth which was exceeded.
        """
        super().__init__('%s exceeds the maximum depth limit of %d'
                         % (action_id, depth_limit))



"""Unit tests for reviewboard.accounts.views.LoginView.

Version Added:
    5.0.5
"""

from reviewboard.site.urlresolvers import local_site_reverse
from reviewboard.testing import TestCase


class LoginViewTests(TestCase):
    """Unit tests for reviewboard.accounts.views.LoginView.

    Version Added:
        5.0.5
    """

    fixtures = ['test_users']

    def test_login_with_redirect(self) -> None:
        """Testing LoginView GET with a local redirect URL"""
        rsp = self.client.get(
            local_site_reverse('login'),
            {
                'next': '/users?foo=1&bar=baz'
            })
        context = rsp.context

        self.assertEqual(context['next'], '/users?foo=1&bar=baz')

    def test_get_client_web_login(self) -> None:
        """Testing LoginView GET sets the redirect field to the client web
        login page when the request indicates the client web login flow and the
        flow is enabled
        """
        settings = {
            'client_web_login': True,
        }
        client_login_url = local_site_reverse('client-login')

        with self.siteconfig_settings(settings):
            rsp = self.client.get(
                local_site_reverse('login'),
                {
                    'client-name': 'TestClient',
                    'client-url': 'http://localhost:8080/test/',
                })

        context = rsp.context

        self.assertEqual(context['client_name'], 'TestClient')
        self.assertEqual(context['client_url'], 'http://localhost:8080/test/')
        self.assertEqual(context['next'],
                         (f'{client_login_url}?client-name=TestClient'
                          '&client-url=http://localhost:8080/test/'))

    def test_get_client_web_login_with_redirect(self) -> None:
        """Testing LoginView GET with the client web login flow encodes
        and passes along a redirect URL if one was given
        """
        settings = {
            'client_web_login': True,
        }
        client_login_url = local_site_reverse('client-login')

        with self.siteconfig_settings(settings):
            rsp = self.client.get(
                local_site_reverse('login'),
                {
                    'client-name': 'TestClient',
                    'client-url': 'http://localhost:8080/test/',
                    'next': 'http://localhost:8080/page?foo=1',
                })

        context = rsp.context

        self.assertEqual(context['client_name'], 'TestClient')
        self.assertEqual(context['client_url'], 'http://localhost:8080/test/')
        self.assertEqual(
            context['next'],
            (f'{client_login_url}?client-name=TestClient'
             '&client-url=http://localhost:8080/test/'
             '&next=http%3A//localhost%3A8080/page%3Ffoo%3D1'))

    def test_get_client_web_login_logged_in(self) -> None:
        """Testing LoginView GET redirects to the client web login
        confirmation page when the request indicates the client web login
        flow and the flow is enabled
        """
        settings = {
            'client_web_login': True,
        }
        client_login_confirm_url = local_site_reverse('client-login-confirm')

        self.client.login(username='doc', password='doc')

        with self.siteconfig_settings(settings):
            rsp = self.client.get(
                local_site_reverse('login'),
                {
                    'client-name': 'TestClient',
                    'client-url': 'http://localhost:8080/test/',
                })

        self.assertRedirects(
            rsp,
            (f'{client_login_confirm_url}?client-name=TestClient'
             '&client-url=http://localhost:8080/test/'))

    def test_get_client_web_login_false(self) -> None:
        """Testing LoginView GET does not set the redirect field to the
        client web login page when the request indicates the client web login
        flow and the flow is not enabled
        """
        settings = {
            'client_web_login': False,
        }

        client_login_url = local_site_reverse('client-login')

        with self.siteconfig_settings(settings):
            rsp = self.client.get(
                local_site_reverse('login'),
                {
                    'client-name': 'TestClient',
                    'client-url': 'http://localhost:8080/test/',
                })

        context = rsp.context

        self.assertNotIn('client_name', context)
        self.assertNotIn('client_url', context)
        self.assertNotEqual(context['next'],
                            (f'{client_login_url}?client-name=TestClient'
                             '&client-url=http://localhost:8080/test/'))

    def test_get_with_redirect_url(self) -> None:
        """Testing LoginView GET with a next= URL"""
        login_url = local_site_reverse('login')

        rsp = self.client.get(
            login_url,
            {
                'next': '/test/',
            })

        self.assertEqual(rsp.context['next'], '/test/')

    def test_get_with_redirect_login_loop(self) -> None:
        """Testing LoginView GET breaks a redirect login loop to the same login
        URL
        """
        login_url = local_site_reverse('login')

        rsp = self.client.get(
            login_url,
            {
                'next': login_url,
            })

        self.assertEqual(rsp.context['next'], '')

    def test_get_with_redirect_logout_loop(self) -> None:
        """Testing LoginView GET breaks a redirect login -> logout loop"""
        rsp = self.client.get(
            local_site_reverse('login'),
            {
                'next': local_site_reverse('logout'),
            })

        self.assertEqual(rsp.context['next'], '')

    def test_get_with_authenticated(self) -> None:
        """Testing LoginView GET when authenticated"""
        client = self.client
        client.login(username='doc',
                     password='doc')

        rsp = client.get(local_site_reverse('login'))
        self.assertRedirects(rsp,
                             local_site_reverse('root'),
                             fetch_redirect_response=False)

    def test_get_with_authenticated_and_next(self) -> None:
        """Testing LoginView GET when authenticated and custom next= URL"""
        client = self.client
        client.login(username='doc',
                     password='doc')

        rsp = client.get(
            local_site_reverse('login'),
            {
                'next': '/test/',
            })

        self.assertRedirects(rsp,
                             '/test/',
                             fetch_redirect_response=False)

    def test_get_with_authenticated_and_login_loop(self) -> None:
        """Testing LoginView GET when authenticated and login loop"""
        client = self.client
        client.login(username='doc',
                     password='doc')

        login_url = local_site_reverse('login')

        rsp = client.get(
            login_url,
            {
                'next': login_url,
            })

        self.assertRedirects(rsp,
                             local_site_reverse('root'),
                             fetch_redirect_response=False)

    def test_get_with_authenticated_and_logout_loop(self) -> None:
        """Testing LoginView GET when authenticated and logout loop"""
        client = self.client
        client.login(username='doc',
                     password='doc')

        rsp = client.get(
            local_site_reverse('login'),
            {
                'next': local_site_reverse('logout'),
            })

        self.assertRedirects(rsp,
                             local_site_reverse('root'),
                             fetch_redirect_response=False)



"""Extra user detail introspection and representation.

Version Added:
    8.0
"""

from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import ClassVar, Optional, TYPE_CHECKING

from django.utils.html import format_html, format_html_join
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _

from reviewboard.accounts.service_accounts import service_account_registry
from reviewboard.registries.registry import OrderedRegistry

if TYPE_CHECKING:
    from collections.abc import Iterator

    from django.contrib.auth.models import User
    from django.http import HttpRequest
    from django.utils.safestring import SafeString

    from reviewboard.site.models import LocalSite


logger = logging.getLogger(__name__)


@dataclass
class UserBadge:
    """A badge shown alongside a user's name.

    Badges can represent some detail about a user. All badges are provided
    by extensions.

    Version Added:
        8.0
    """

    ######################
    # Instance variables #
    ######################

    #: The visual label shown on the badge.
    label: str

    #: The user this badge belongs to.
    user: User

    #: An optional CSS class (or classes) used for the badge.
    #:
    #: This allows an extension to provide custom styling for a badge, such
    #: as custom colors.
    #:
    #: Multiple CSS classes can be provided by separating them with spaces.
    css_class: Optional[str] = None

    def render_to_string(self) -> SafeString:
        """Render the badge to an HTML string.

        All badges use a ``rb-c-user-badge`` CSS class, along with any
        CSS classes provided in :py:attr:`css_class`.

        The label for the badge will reflect :py:attr:`label`.

        Returns:
            django.utils.safestring.SafeString:
            The HTML for the badge.
        """
        try:
            css_classes: list[str] = [
                'rb-c-user-badge',
            ]

            css_class = self.css_class

            if css_class:
                css_classes.append(css_class)

            attrs: list[tuple[str, str]] = [
                ('class', ' '.join(css_classes)),
                ('role', 'listitem'),
            ]

            return format_html(
                '<span {attrs}>{label}</span>',
                attrs=format_html_join(' ', '{}="{}"', attrs),
                label=self.label)
        except Exception as e:
            logger.exception(
                'Unexpected error when rendering user badge %r: %s',
                self, e)

            return mark_safe('')


class BaseUserDetailsProvider:
    """Base class for a provider for additional user details.

    This enables extensions to provide additional information about users,
    for specialized rendering and logic.

    Version Added:
        8.0
    """

    #: The unique ID of the user details provider.
    user_details_provider_id: ClassVar[str]

    def get_user_badges(
        self,
        user: User,
        *,
        local_site: Optional[LocalSite],
        request: Optional[HttpRequest] = None,
    ) -> Iterator[UserBadge]:
        """Return a list of badges to display for a user.

        Badges provide a visual indicator about some aspect of a user, for
        display purposes.

        Args:
            user (django.contrib.auth.models.User):
                The user the badges must correspond to.

            local_site (reviewboard.site.models.LocalSite):
                The Local Site the badges must correspond to.

            request (django.http.HttpRequest, optional):
                The optional HTTP request from the client.

        Yields:
            UserBadge:
            Each badge to display.
        """
        yield from []


class ServiceAccountDetailsProvider(BaseUserDetailsProvider):
    """A user details provider that identifies service accounts.

    Each registered service account will have a "Service Account" badge beside
    its name, identifying it so it's not confused as a normal account.

    Version Added:
        8.1
    """

    user_details_provider_id = 'service-accounts'

    def get_user_badges(
        self,
        user: User,
        *,
        local_site: LocalSite | None,
        request: (HttpRequest | None) = None,
    ) -> Iterator[UserBadge]:
        """Return all badges for service account users.

        If the user is a registered service account, a "Service Account"
        badge will be displayed.

        Args:
            user (django.contrib.auth.models.User):
                The user the badges must correspond to.

            local_site (reviewboard.site.models.LocalSite):
                The Local Site the badges must correspond to.

            request (django.http.HttpRequest, optional):
                The optional HTTP request from the client.

        Yields:
            UserBadge:
            Each badge to display.
        """
        if service_account_registry.get_for_username(user.username):
            yield UserBadge(
                label=_('Service Account'),
                user=user,
            )


class UserDetailsProviderRegistry(OrderedRegistry[BaseUserDetailsProvider]):
    """A registry of user detail providers.

    Each provider in the registry will be queried when requesting additional
    information on users.

    Version Added:
        8.0
    """

    lookup_attrs = ('user_details_provider_id',)

    def get_user_details_provider(
        self,
        provider_id: str,
    ) -> Optional[BaseUserDetailsProvider]:
        """Return the user details provider for a given ID.

        Args:
            provider_id (str):
                The ID of the user details provider to return.

        Returns:
            UserDetailsProvider:
            The user details provider matching the ID, or ``None`` if not
            found.
        """
        return self.get('user_details_provider_id', provider_id)

    def get_defaults(self) -> Iterator[BaseUserDetailsProvider]:
        """Return default user details providers for the registry.

        Version Added:
            8.1

        Yields:
            BaseUserDetailsProvider:
            Each default user details provider.
        """
        yield ServiceAccountDetailsProvider()


#: The registry managing user details providers.
#:
#: Version Added:
#:     8.0
user_details_provider_registry = UserDetailsProviderRegistry()



"""Admin definitions for the FileAttachment model."""

from __future__ import annotations

from django.contrib import admin
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

from reviewboard.admin import ModelAdmin, admin_site
from reviewboard.attachments.models import FileAttachment


class FileAttachmentAdmin(ModelAdmin):
    """Admin definitions for the FileAttachment model."""

    list_display = ('filename', 'caption', 'mimetype', 'review_request_id',
                    'download')
    list_display_links = ('filename', 'caption')
    search_fields = ('caption', 'mimetype')
    raw_id_fields = ('added_in_filediff', 'local_site', 'user')

    def filename(
        self,
        obj: FileAttachment,
    ) -> str:
        """Return the name of the file.

        Args:
            obj (reviewboard.attachments.models.FileAttachment):
                The file attachment.

        Returns:
            str:
            The name of the file.
        """
        if obj.filename:
            return obj.filename[:80]
        else:
            return '(unnamed)'

    @admin.display(description=_('Review Request ID'))
    def review_request_id(
        self,
        obj: FileAttachment,
    ) -> int:
        """Return the review request ID for this file attachment.

        Args:
            obj (reviewboard.attachments.models.FileAttachment):
                The file attachment.

        Returns:
            int:
            The ID of the linked review request.
        """
        return obj.get_review_request().id

    def download(
        self,
        obj: FileAttachment,
    ) -> str:
        """Return a link to download the file.

        Args:
            obj (reviewboard.attachments.models.FileAttachment):
                The file attachment.

        Returns:
            str:
            A link to download the file
        """
        return format_html(
            '<a href="{}"><span class="rb-icon rb-icon-download"></span></a>',
            obj.file)


admin_site.register(FileAttachment, FileAttachmentAdmin)



"""A hook for checking ACLs on diff files."""

from __future__ import annotations

from djblets.extensions.hooks import ExtensionHook, ExtensionHookPoint


class FileDiffACLHook(ExtensionHook, metaclass=ExtensionHookPoint):
    """A hook for checking ACLs on diff files.

    Extensions can use this hook to connect repository ACLs into the Review
    Board access system. This is provided as an extension hook because
    systems may be deployed in various ways, and SCM usernames may not
    necessarily match Review Board usernames.

    Version Added:
        4.0.5:
        This is experimental in 4.0.x, with plans to make it stable for 5.0.
        The API may change during this time.
    """

    def is_accessible(self, diffset, user, **kwargs):
        """Return whether the given file is accessible by the given user.

        Args:
            diffset (reviewboard.diffviewer.models.DiffSet):
                The diffset containing the file.

            user (django.contrib.auth.models.User):
                The user to check.

            **kwargs (dict, unused):
                Additional keyword arguments for future expansion.

        Returns:
            bool:
            False if the user does not have access to the file. True if the
            user explicitly does have access. None if the extension did not
            check for this diffset or repository (so that other hook points can
            continue).
        """
        raise NotImplementedError



"""Add issue_opened and issue_status to Comment and ScreenshotComment.

Version Added:
    1.6
"""

from django_evolution.mutations import AddField
from django.db import models

MUTATIONS = [
    AddField('Comment', 'issue_opened', models.BooleanField, initial=False),
    AddField('Comment', 'issue_status', models.CharField,
             initial='', max_length=1, null=True, db_index=True),
    AddField('ScreenshotComment', 'issue_opened', models.BooleanField,
             initial=False),
    AddField('ScreenshotComment', 'issue_status', models.CharField,
             initial='', max_length=1, null=True, db_index=True)
]
